from chisubmit.tests.common import cli_test, ChisubmitCLITestCase

class CLIInstructorGrades(ChisubmitCLITestCase):
            
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                     'course1_pa1', 'course1_pa1_registrations_with_submissions',
                     'course1_pa1_grades',
                     'course1_pa2']
                
    @cli_test
    def test_instructor_list_grades(self, runner):
        _, instructors, _, _ = self.create_clients(runner, "admin", instructor_ids=["instructor1"], course_id="cmsc40100")
        
        instructor1 = instructors[0]
        
        result = instructor1.run("instructor grading list-grades --detailed")
        self.assertEquals(result.exit_code, 0)
        
    @cli_test
    def test_instructor_grading_status(self, runner):
        _, instructors, _, _ = self.create_clients(runner, "admin", instructor_ids=["instructor1"], course_id="cmsc40100")
        
        instructor1 = instructors[0]
        
        result = instructor1.run("instructor grading show-grading-status --by-grader", ["pa1"])
        self.assertEquals(result.exit_code, 0)


from chisubmit.tests.common import cli_test, ChisubmitCLITestCase
from chisubmit.common.utils import get_datetime_now_utc
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL

from datetime import timedelta
from chisubmit.backend.api.models import Course, get_user_by_username

class CLICompleteWorkflowExtensionsPerStudent(ChisubmitCLITestCase):
        
    fixtures = ['admin_user']
    
    @cli_test
    def test_complete_with_extensions_per_student(self, runner):
        course_id = u"cmsc40300"
        course_name = u"Foobarmentals of Foobar II"

        admin_id = u"admin"
        instructor_ids = [u"instructor"]
        grader_ids= [u"grader"]
        student_ids = [u"student1", u"student2", u"student3", u"student4"]
        
        all_users = instructor_ids + grader_ids + student_ids
        
        admin, instructors, graders, students = self.create_clients(runner, admin_id, instructor_ids, grader_ids, student_ids, course_id, verbose = True)
        self.create_users(admin, all_users)
                
        self.create_course(admin, course_id, course_name)

        course = Course.get_by_course_id(course_id)
        self.assertIsNotNone(course)
        self.assertEquals(course.name, course_name)

        result = admin.run("admin course set-attribute %s default_extensions 3" % (course_id))
        self.assertEquals(result.exit_code, 0)
        
        result = admin.run("admin course set-attribute %s extension_policy per-student" % (course_id))
        self.assertEquals(result.exit_code, 0)
        
        self.add_users_to_course(admin, course_id, instructors, graders, students)
        
        deadline = get_datetime_now_utc() - timedelta(hours=23)
        deadline = deadline.replace(tzinfo=None).isoformat(sep=" ")        

        result = instructors[0].run("instructor assignment add", 
                                    ["pa1", "Programming Assignment 1", deadline])
        self.assertEquals(result.exit_code, 0)

        result = instructors[0].run("instructor assignment set-attribute", 
                                    ["pa1", "max_students", "2"])
        self.assertEquals(result.exit_code, 0)

        deadline = get_datetime_now_utc() - timedelta(hours=47)
        deadline = deadline.isoformat(sep=" ")

        result = instructors[0].run("instructor assignment add", 
                                    ["pa2", "Programming Assignment 2", deadline])
        self.assertEquals(result.exit_code, 0)

        result = instructors[0].run("instructor assignment set-attribute", 
                                    ["pa2", "max_students", "2"])
        self.assertEquals(result.exit_code, 0)

        deadline = get_datetime_now_utc() - timedelta(hours=47)
        deadline = deadline.isoformat(sep=" ")

        result = instructors[0].run("instructor assignment add", 
                                    ["pa3", "Programming Assignment 3", deadline])
        self.assertEquals(result.exit_code, 0)

        result = instructors[0].run("instructor assignment set-attribute", 
                                    ["pa3", "max_students", "2"])
        self.assertEquals(result.exit_code, 0)

        deadline = get_datetime_now_utc() - timedelta(hours=23)
        deadline = deadline.replace(tzinfo=None).isoformat(sep=" ")        

        result = instructors[0].run("instructor assignment add", 
                                    ["pa4", "Programming Assignment 4", deadline])
        self.assertEquals(result.exit_code, 0)

        result = instructors[0].run("instructor assignment set-attribute", 
                                    ["pa4", "max_students", "2"])
        self.assertEquals(result.exit_code, 0)

        deadline = get_datetime_now_utc() + timedelta(hours=2)
        deadline = deadline.replace(tzinfo=None).isoformat(sep=" ")        

        result = instructors[0].run("instructor assignment add", 
                                    ["pa5", "Programming Assignment 5", deadline])
        self.assertEquals(result.exit_code, 0)

        result = instructors[0].run("instructor assignment set-attribute", 
                                    ["pa5", "max_students", "2"])
        self.assertEquals(result.exit_code, 0)
        
        
        teams = [u"student1-student2", 
                 u"student3-student4",
                 u"student1-student3",
                 u"student2-student4"]        

        students_team = [ (students[0], students[1]),
                          (students[2], students[3]),
                          (students[0], students[2]),
                          (students[1], students[3])]
        

        self.register_team(students_team[0], teams[0], "pa1", course_id)
        self.register_team(students_team[1], teams[1], "pa2", course_id)
        
        team_git_paths, team_git_repos, team_commits = self.create_team_repos(admin, course_id, teams[0:2], students_team[0:2])

        self.register_team(students_team[2], teams[2], "pa3", course_id)
        self.register_team(students_team[2], teams[2], "pa4", course_id)
        self.register_team(students_team[2], teams[2], "pa5", course_id)
        self.register_team(students_team[3], teams[3], "pa3", course_id)
        self.register_team(students_team[3], teams[3], "pa4", course_id)
        self.register_team(students_team[3], teams[3], "pa5", course_id)
        
        x, y, z = self.create_team_repos(admin, course_id, teams[2:4], students_team[2:4])
        
        team_git_paths += x
        team_git_repos += y
        team_commits += z
        
        for s in students:
            result = s.run("student course show-extensions")
            self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)        
        
        # Team 0 submits with one extension to pa1
        # Student 0 and 1 now have 2 extensions left each
        result = students_team[0][0].run("student assignment submit", 
                                         ["pa1", "--yes"])        
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)
        
        s1 = course.get_student(get_user_by_username(students_team[0][0].user_id))
        s2 = course.get_student(get_user_by_username(students_team[0][1].user_id))
        t = course.get_team(teams[0])
        
        self.assertEqual(t.get_extensions_available(), 2)
        self.assertEqual(s1.get_extensions_available(), 2)
        self.assertEqual(s2.get_extensions_available(), 2)

        for s in students:
            result = s.run("student course show-extensions")
            self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)
        
        # Team 1 submits with two extensions to pa2
        # Student 2 and 3 now have 1 extension left each
        result = students_team[1][0].run("student assignment submit", 
                                         ["pa2", "--yes"])        
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        s1 = course.get_student(get_user_by_username(students_team[1][0].user_id))
        s2 = course.get_student(get_user_by_username(students_team[1][1].user_id))
        t = course.get_team(teams[1])
        
        self.assertEqual(t.get_extensions_available(), 1)
        self.assertEqual(s1.get_extensions_available(), 1)
        self.assertEqual(s2.get_extensions_available(), 1)

        for s in students:
            result = s.run("student course show-extensions")
            self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        result = students_team[0][0].run("student team show", [teams[0]])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        result = students_team[1][0].run("student team show", [teams[1]])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)


        # Now we have two new teams:
        #
        # Team 2 with Student 0 and 2, with 2 and 1 extensions left respectively.
        # This team can only submit to projects requiring one extension
        # 
        # Team 3 with Student 1 and 3, with 2 and 1 extensions left respectively.
        # This team can only submit to projects requiring one extension

        # Team 2 tries to submit to pa3, but they are denied because they would 
        # need two extensions to submit.
        result = students_team[2][0].run("student assignment submit", 
                                         ["pa3", "--yes"])        
        self.assertEquals(result.exit_code, CHISUBMIT_FAIL)

        s1 = course.get_student(get_user_by_username(students_team[2][0].user_id))
        s2 = course.get_student(get_user_by_username(students_team[2][1].user_id))
        t = course.get_team(teams[2])
        
        self.assertEqual(t.get_extensions_available(), 1)
        self.assertEqual(s1.get_extensions_available(), 2)
        self.assertEqual(s2.get_extensions_available(), 1)


        # Team 2 tries submitting to pa4 and it goes through because it only
        # requires one extensions
        result = students_team[2][0].run("student assignment submit", 
                                         ["pa4", "--yes"])        
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        s1 = course.get_student(get_user_by_username(students_team[2][0].user_id))
        s2 = course.get_student(get_user_by_username(students_team[2][1].user_id))
        t = course.get_team(teams[2])
        
        self.assertEqual(t.get_extensions_available(), 0)
        self.assertEqual(s1.get_extensions_available(), 1)
        self.assertEqual(s2.get_extensions_available(), 0)

        for s in students:
            result = s.run("student course show-extensions")
            self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)                        

        # Team 2 tries submitting to pa5 with zero extensions and is accepted
        result = students_team[2][0].run("student assignment submit", 
                                         ["pa5", "--yes"])        
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)
                        
        s1 = course.get_student(get_user_by_username(students_team[2][0].user_id))
        s2 = course.get_student(get_user_by_username(students_team[2][1].user_id))
        t = course.get_team(teams[2])
        
        self.assertEqual(t.get_extensions_available(), 0)
        self.assertEqual(s1.get_extensions_available(), 1)
        self.assertEqual(s2.get_extensions_available(), 0)

        for s in students:
            result = s.run("student course show-extensions")
            self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)                        

        # Team 3 tries submitting to pa5 and is accepted (doesn't require extensions)
        result = students_team[3][0].run("student assignment submit", 
                                         ["pa5", "--yes"])        
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        s1 = course.get_student(get_user_by_username(students_team[3][0].user_id))
        s2 = course.get_student(get_user_by_username(students_team[3][1].user_id))
        t = course.get_team(teams[3])
        
        self.assertEqual(t.get_extensions_available(), 1)
        self.assertEqual(s1.get_extensions_available(), 2)
        self.assertEqual(s2.get_extensions_available(), 1)

        for team, student_team in zip(teams, students_team):
            result = student_team[0].run("student team show", [team])
            self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        for s in students:
            result = s.run("student course show-extensions")
            self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)        
        
        



import click

from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from chisubmit.repos.grading import GradingGitRepo
from chisubmit.rubric import RubricFile, ChisubmitRubricException
from chisubmit.cli.common import create_grading_repos,\
    gradingrepo_push_grading_branch, gradingrepo_pull_grading_branch,\
    get_assignment_or_exit, get_teams_registrations, get_team_or_exit,\
    get_assignment_registration_or_exit, get_grader_or_exit,\
    catch_chisubmit_exceptions, require_local_config
from chisubmit.cli.common import pass_course
from chisubmit.common.utils import create_connection

import csv
import operator
import random
import itertools
import os.path
import yaml
from chisubmit.client.exceptions import UnknownObjectException
import math

@click.group(name="grading")
@click.pass_context
def instructor_grading(ctx):
    pass


@click.command(name="set-grade")
@click.argument('team_id', type=str)
@click.argument('assignment_id', type=str)
@click.argument('rubric_component_description', type=str)
@click.argument('points', type=float)
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_set_grade(ctx, course, team_id, assignment_id, rubric_component_description, points):   
    assignment = get_assignment_or_exit(ctx, course, assignment_id)
    team = get_team_or_exit(ctx, course, team_id)
    registration = get_assignment_registration_or_exit(ctx, team, assignment_id)
     
    rubric_components = assignment.get_rubric_components()
    
    rubric_component = [rc for rc in rubric_components if rc.description == rubric_component_description]
    
    if len(rubric_component) == 0:
        print "No such rubric component in %s: %s" % (assignment_id, rubric_component_description)
        ctx.exit(CHISUBMIT_FAIL)
    elif len(rubric_component) > 1:
        print "ERROR: Server returned more than one rubric component for '%s' in %s" % (rubric_component_description, assignment_id)
        print "       This should not happen. Please contact the chisubmit administrator."
        ctx.exit(CHISUBMIT_FAIL)
    else:
        rc = rubric_component[0]
        
        try:
            registration.set_grade(rc, points)
            ctx.exit(CHISUBMIT_SUCCESS)
        except ValueError, ve:
            print ve.message
            ctx.exit(CHISUBMIT_FAIL)
            


@click.command(name="load-grades")
@click.argument('assignment_id', type=str)
@click.argument('grade_component_id', type=str)
@click.argument('csv_file', type=click.File('rb'))
@click.argument('csv_team_column', type=str)
@click.argument('csv_grade_column', type=str)
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_load_grades(ctx, course, assignment_id, grade_component_id, csv_file, csv_team_column, csv_grade_column):   
    assignment = course.get_assignment(assignment_id)
    if assignment is None:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)

    grade_component = assignment.get_grade_component(grade_component_id)
    if not grade_component:
        print "Assignment %s does not have a grade component '%s'" % (assignment.assignment_id, grade_component)
        ctx.exit(CHISUBMIT_FAIL)
        
    csvf = csv.DictReader(csv_file)
            
    if csv_team_column not in csvf.fieldnames:
        print "CSV file %s does not have a '%s' column" % (csv_file, csv_team_column)
        ctx.exit(CHISUBMIT_FAIL)
        
    if csv_grade_column not in csvf.fieldnames:
        print "CSV file %s does not have a '%s' column" % (csv_file, csv_grade_column)
        ctx.exit(CHISUBMIT_FAIL)
            
    for entry in csvf:
        team_id = entry[csv_team_column]
        
        team = course.get_team(team_id)
        if team is None:
            print "%-40s SKIPPING. Not a team in course %s" % (team_id, course.id)
            continue

        ta = team.get_assignment(assignment_id)
        if ta is None:
            print "%-40s SKIPPING. Not registered for assignment %s" % (team_id, assignment_id)
            continue
        
        if ta.submitted_at is None:
            print "%-40s SKIPPING. Has not submitted assignment %s yet" % (team_id, assignment_id)
            continue
    
        grade = entry[csv_grade_column]
    
        if grade is None or grade == "":
            grade = 0
        else:
            grade = float(grade)

        if grade < 0 or grade > grade_component.points:
            print "%-40s SKIPPING. Invalid grade value %.2f (%s must be 0 <= x <= %.2f)" % (team_id, grade, grade_component_id, grade_component.points)
            continue
        
        try:
            team.set_assignment_grade(assignment_id, grade_component.id, grade)
            print "%-40s %s <- %.2f" % (team_id, grade_component_id, grade)
        except Exception:
            raise
            

@click.command(name="add-conflict")
@click.argument('grader_id', type=str)
@click.argument('student_id', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_add_conflict(ctx, course, grader_id, student_id):
    grader = course.get_grader(grader_id)
    if not grader:
        print "Grader %s does not exist" % grader_id
        ctx.exit(CHISUBMIT_FAIL)

    student = course.get_student(student_id)
    if not student:
        print "Student %s does not exist" % student_id
        ctx.exit(CHISUBMIT_FAIL)

    if student in grader.conflicts:
        print "Student %s is already listed as a conflict for grader %s" % (student.id, grader.id)

    grader.conflicts.append(student)

    return CHISUBMIT_SUCCESS

@click.command(name="list-grades")
@click.option('--detailed', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_list_grades(ctx, course, detailed):
    students = [s for s in course.get_students() if not s.dropped]
    assignments = course.get_assignments(include_rubric=True)

    students.sort(key=operator.attrgetter("user.last_name"))
    assignments.sort(key=operator.attrgetter("deadline"))

    student_grades = dict([(s.user.username,dict([(a.assignment_id,{}) for a in assignments])) for s in students])

    teams = course.get_teams(include_students=True, include_assignments=True, include_grades=True)

    for team in teams:
        registrations = team.get_assignment_registrations()
        for registration in registrations:
            assignment_id = registration.assignment.assignment_id
            for student in team.get_team_members():
                student_id = student.username
                if student_grades.has_key(student_id):
                    g = student_grades[student_id][assignment_id]
                    for grade in registration.get_grades():
                        rc_description = grade.rubric_component.description
                        if g.has_key(rc_description):
                            print "Warning: %s already has a grade for rubric component '%s'" % (student_id, rc_description)
                        else:
                            g[rc_description] = grade.points
                    g["__PENALTIES"] = registration.get_total_penalties()
                    g["__BONUSES"] = registration.get_total_bonuses()                    
                    g["__TOTAL"] = registration.get_total_grade()

    fields = ["Username","Last Name","First Name"]
    for assignment in assignments:
        if detailed:
            rubric_components = assignment.get_rubric_components()
            fields += ["%s - %s" % (assignment.assignment_id, rc.description) for rc in rubric_components]
            fields.append("%s - Penalties" % assignment.assignment_id)
            fields.append("%s - Bonuses" % assignment.assignment_id) 
            fields.append("%s - Total" % assignment.assignment_id)
        else:
            fields.append("%s" % assignment.assignment_id)

    print ",".join(fields)

    for student in students:
        fields = [student.user.username, student.user.last_name, student.user.first_name]
        for assignment in assignments:
            grades = student_grades[student.user.username][assignment.assignment_id]
            if detailed:
                rubric_components = assignment.get_rubric_components()
                for rc in rubric_components:
                    if not grades.has_key(rc.description):
                        fields.append("")
                    else:
                        fields.append(str(grades[rc.description]))
                if len(grades) == 0:
                    fields += ["","",""]
                else:
                    fields.append(str(grades["__PENALTIES"]))
                    fields.append(str(grades["__BONUSES"]))
                    fields.append(str(grades["__TOTAL"]))
            else:
                if len(grades) == 0:
                    fields.append("")
                else:
                    fields.append(str(grades["__TOTAL"]))

        print ",".join(fields)


@click.command(name="assign-graders")
@click.argument('assignment_id', type=str)
@click.option('--from-assignment', type=str)
@click.option('--avoid-assignment', type=str)
@click.option('--grader-file', type=click.File('r'))
@click.option('--dry-run', is_flag=True)
@click.option('--reset', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_assign_graders(ctx, course, assignment_id, from_assignment, avoid_assignment, grader_file, dry_run, reset):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    if from_assignment is not None:
        from_assignment = get_assignment_or_exit(ctx, course, from_assignment)

    if avoid_assignment is not None:
        avoid_assignment = get_assignment_or_exit(ctx, course, avoid_assignment)

    if reset and from_assignment is not None:
        print "--reset and --from_assignment are mutually exclusive"
        ctx.exit(CHISUBMIT_FAIL)

    if avoid_assignment is not None and from_assignment is not None:
        print "--avoid_assignment and --from_assignment are mutually exclusive"
        ctx.exit(CHISUBMIT_FAIL)

    if grader_file is not None:
        grader_workload = yaml.load(grader_file)
        graders = []
        
        for username in grader_workload:
            try:
                grader = course.get_grader(username)
                graders.append(grader)
            except UnknownObjectException:
                print "No such grader: %s" % username
                ctx.exit(CHISUBMIT_FAIL)
    else:
        graders = course.get_graders()
        grader_workload = {g.user.username: "remainder" for g in graders}

    if len(graders) == 0:
        print "ERROR: No graders."
        ctx.exit(CHISUBMIT_FAIL)
            
    teams_registrations = get_teams_registrations(course, assignment)
    teams = teams_registrations.keys()

    n_teams = len(teams)
    teams_per_grader = {}
    teams_per_grader_assigned = dict([(g.user.username, 0) for g in graders])    
    assigned = 0
    graders_assigned_remainder = []
    for username, workload in grader_workload.items():
        if workload != "remainder":
            if not isinstance(workload, int) or workload <= 0:
                print "Invalid workload for grader '%s': %s" % (username, workload)
            teams_per_grader[username] = workload
            assigned += workload
        else:
            teams_per_grader[username] = None
            graders_assigned_remainder.append(username)
                        
    min_teams_per_grader = (n_teams - assigned) / len(graders_assigned_remainder)
    extra_teams = (n_teams - assigned) % len(graders_assigned_remainder)

    for username in teams_per_grader:
        if teams_per_grader[username] is None:
            teams_per_grader[username] = min_teams_per_grader
            
    random.seed(course.course_id + assignment_id)
    random.shuffle(graders_assigned_remainder)

    # so many graders in this course that some will end up expecting zero
    # teams to grade. Make sure they are able to get at least one.
    for username in graders_assigned_remainder[:extra_teams]:
        teams_per_grader[username] += 1
    
    assert sum([v for v in teams_per_grader.values()]) == n_teams

    team_grader = {t.team_id: None for t in teams_registrations.keys()}

    if from_assignment is not None:
        from_assignment_registrations = get_teams_registrations(course, from_assignment)
        
        common_teams = [t for t in teams if teams_registrations.has_key(t) and from_assignment_registrations.has_key(t)]
        for t in common_teams:
            registration = from_assignment_registrations[t]

            # try to assign the same grader that would grade the same team's other assignment
            grader = registration.grader
            if grader is not None and teams_per_grader[grader.user.username] > 0:
                team_grader[t.team_id] = grader.user.username 
                teams_per_grader[grader.user.username] -= 1
                teams_per_grader_assigned[grader.user.username] += 1

    if avoid_assignment is not None:
        avoid_assignment_registrations = get_teams_registrations(course, avoid_assignment)

    if not reset:
        for t in teams:
            if teams_registrations[t].grader is None:
                team_grader[t.team_id] = None
            else:
                grader_id = teams_registrations[t].grader.user.username
                team_grader[t.team_id] = grader_id
                teams_per_grader[grader_id] = max(0, teams_per_grader[grader_id] - 1)
                teams_per_grader_assigned[grader_id] += 1

    not_ready_for_grading = []
    ta_avoid = {}
    graders_cycle = itertools.cycle(graders)
    for team, registration in teams_registrations.items():
        if team_grader[team.team_id] is not None:
            continue

        if not registration.is_ready_for_grading():
            not_ready_for_grading.append(team.team_id)
            continue

        for g in graders_cycle:
            if teams_per_grader[g.user.username] == 0:
                continue
            else:
                if avoid_assignment is not None:
                    avoid_registration = ta_avoid.setdefault(team.team_id, avoid_assignment_registrations.get(team))
                    if avoid_registration is not None and avoid_registration.grader.user.username == grader.user.username:
                        continue
                
                valid = True
                for tm in team.get_team_members():
                    conflicts = g.get_conflicts()
                    if tm.username in conflicts:
                        valid = False
                        break

                if valid:
                    team_grader[team.team_id] = g.user.username 
                    teams_per_grader[g.user.username] -= 1
                    teams_per_grader_assigned[g.user.username] += 1                        
                    break                    
    
    for team, registration in teams_registrations.items():
        if team_grader[team.team_id] is None:
            if team.team_id not in not_ready_for_grading:
                print "Team %s has no grader" % (team.team_id)
            else:
                print "Team %s's submission isn't ready for grading yet" % (team.team_id)
        else:
            if not dry_run:
                if registration.grader_username != team_grader[team.team_id]:
                    registration.grader_username = team_grader[team.team_id]
            else:
                print "%s: %s" % (team.team_id, team_grader[team.team_id])
    print 
    for grader_id, assigned in teams_per_grader_assigned.items():
        if teams_per_grader[grader_id] != 0:
            print grader_id, assigned, "(still needs to be assigned %i more assignments)" % (teams_per_grader[grader_id])
        else:
            print grader_id, assigned

    return CHISUBMIT_SUCCESS


@click.command(name="list-grader-assignments")
@click.argument('assignment_id', type=str)
@click.option('--grader-id', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_list_grader_assignments(ctx, course, assignment_id, grader_id):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    if grader_id is not None:
        grader = get_grader_or_exit(ctx, course, grader_id)
    else:
        grader = None

    teams_registrations = get_teams_registrations(course, assignment)
    
    teams = teams_registrations.keys()
    teams.sort(key=operator.attrgetter("team_id"))

    for t in teams:
        registration = teams_registrations[t]
        if grader is None:
            if registration.grader is None:
                grader_str = "<no-grader-assigned>"
            else:
                grader_str = registration.grader.user.username
            print t.team_id, grader_str
        else:
            if grader == registration.grader.user.username:
                print t.team_id

    return CHISUBMIT_SUCCESS


@click.command(name="list-submissions")
@click.argument('assignment_id', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_list_submissions(ctx, course, assignment_id):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    teams_registrations = get_teams_registrations(course, assignment)
    teams = sorted(teams_registrations.keys(), key=operator.attrgetter("team_id"))

    conn = create_connection(course, ctx.obj['config'])
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)

    for team in teams:
        registration = teams_registrations[team]

        if registration.final_submission is None:
            print "%25s NOT SUBMITTED" % team.team_id
        else:
            submission_commit = conn.get_commit(course, team, registration.final_submission.commit_sha)
            if submission_commit is not None:
                if registration.is_ready_for_grading():
                    print "%25s SUBMITTED (READY for grading)" % team.team_id
                else:
                    print "%25s SUBMITTED (NOT READY for grading)" % team.team_id
            else:
                print "%25s ERROR: Submitted but commit %s not found in repository" % (team.team_id, registration.final_submission.commit_sha)
    
    return CHISUBMIT_SUCCESS


def print_grades_stats(grades):
    grades.sort()
    avg = sum(grades) / len(grades)
    stdev = math.sqrt(sum([(x-avg)**2 for x in grades])/len(grades))
    print
    print "     Average grade: %.2f" % avg
    print "Standard deviation: %.2f" % stdev
    print
    print "               Max: %.2f" % grades[-1]
    print "    Upper Quartile: %.2f" % grades[ int(len(grades)*0.75) ]
    print "            Median: %.2f" % grades[ int(len(grades)*0.5) ]
    print "    Lower Quartile: %.2f" % grades[ int(len(grades)*0.25) ]
    print "               Min: %.2f" % grades[0]
    print



@click.command(name="show-grading-status")
@click.argument('assignment_id', type=str)
@click.option('--by-grader', is_flag=True)
@click.option('--include-diff-urls', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_show_grading_status(ctx, course, assignment_id, by_grader, include_diff_urls):
    assignment = get_assignment_or_exit(ctx, course, assignment_id, include_rubric = True)
    rubric_components = assignment.get_rubric_components()

    teams_registrations = get_teams_registrations(course, assignment, include_grades = True)
    teams = sorted(teams_registrations.keys(), key=operator.attrgetter("team_id"))
    
    team_status = []
    graders = set()
    
    for team in teams:
        registration = teams_registrations[team]
        grades = registration.get_grades()
        
        grade_rc_ids = [g.rubric_component_id for g in grades]

        if registration.grader is None:
            grader_str = "<no grader assigned>"
        else:
            grader_str = registration.grader.user.username
            
        graders.add(grader_str)

        has_some = False
        has_all = True
        for rc in rubric_components:
            if rc.id in grade_rc_ids:
                has_some = True
            else:
                has_all = False
                        
        if not has_some:
            grading_status = "NOT GRADED"
        elif has_all:
            grading_status = "GRADED"
        else:
            grading_status = "PARTIALLY GRADED"

        total_grade = registration.get_total_grade()
            
        if include_diff_urls and has_some:
            commit_sha = registration.final_submission.commit_sha[:8]
            diff_url = "https://mit.cs.uchicago.edu/%s-staging/%s/compare/%s...%s-grading" % (course.course_id, team.team_id, commit_sha, assignment.assignment_id)
        else:
            diff_url = ""
            
        team_status.append((team.team_id, grader_str, total_grade, diff_url, grading_status))

    if not by_grader:
        for team, grader, total_grade, diff_url, status in team_status:
            print "%-40s %-20s %-20.2f %10s  %s" % (team, status, total_grade, grader, diff_url)
    else:
        all_grades = []
        for grader in sorted(list(graders)):
            print grader
            print "-" * len(grader)
            
            grades = []
            
            team_status_grader = [ts for ts in team_status if ts[1] == grader]
            
            for team, _, total_grade, diff_url, status in team_status_grader:
                if status == "NOT GRADED":
                    print "%-40s %s  %s" % (team, status, diff_url)
                else:
                    print "%-40s %s %8.2f  %s" % (team, status, total_grade, diff_url)
                    if status == "GRADED":
                        grades.append(total_grade)
                        
            if len(grades) > 0:
                all_grades += grades
                print_grades_stats(grades)
            print

        if len(all_grades) > 0:
            print "TOTAL"
            print "-----"
            print_grades_stats(all_grades)

    return CHISUBMIT_SUCCESS

@click.command(name="create-grading-repos")
@click.argument('assignment_id', type=str)
@click.option('--all-teams', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_create_grading_repos(ctx, course, assignment_id, all_teams):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)
    
    teams_registrations = get_teams_registrations(course, assignment, only_ready_for_grading=not all_teams)

    repos = create_grading_repos(ctx.obj['config'], course, assignment, teams_registrations)

    if repos == None:
        ctx.exit(CHISUBMIT_FAIL)

    return CHISUBMIT_SUCCESS


@click.command(name="create-grading-branches")
@click.argument('assignment_id', type=str)
@click.option('--all-teams', is_flag=True)
@click.option('--only', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_create_grading_branches(ctx, course, assignment_id, all_teams, only):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    teams_registrations = get_teams_registrations(course, assignment, only = only, only_ready_for_grading=not all_teams)
    teams = sorted(teams_registrations.keys(), key=operator.attrgetter("team_id"))

    if len(teams_registrations) == 0:
        ctx.exit(CHISUBMIT_FAIL)

    for team in teams:
        registration = teams_registrations[team]
        repo = GradingGitRepo.get_grading_repo(ctx.obj['config'], course, team, registration)

        if repo is None:
            print "%s does not have a grading repository" % team.team_id
            continue
        
        if registration.final_submission is None:
            submitted_at = None
        else:
            submitted_at = registration.final_submission.submitted_at
            
        if submitted_at is None:
            print "Skipping grading branch. %s has not submitted." % team.team_id
        elif repo.has_grading_branch():
            print "Skipping grading branch. %s already has a grading branch." % team.team_id
        else:
            repo.create_grading_branch()
            print "Created grading branch for %s" % team.team_id

    return CHISUBMIT_SUCCESS


@click.command(name="push-grading-branches")
@click.argument('assignment_id', type=str)
@click.option('--to-staging', is_flag=True)
@click.option('--to-students', is_flag=True)
@click.option('--all-teams', is_flag=True)
@click.option('--only', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_push_grading_branches(ctx, course, assignment_id, to_staging, to_students, all_teams, only):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)
    
    teams_registrations = get_teams_registrations(course, assignment, only = only, only_ready_for_grading=not all_teams)
    teams = sorted(teams_registrations.keys(), key=operator.attrgetter("team_id"))

    for team in teams:
        registration = teams_registrations[team]
        print ("Pushing grading branch for team %s... " % team.team_id),
        gradingrepo_push_grading_branch(ctx.obj['config'], course, team, registration, to_staging = to_staging, to_students = to_students)
        print "done."

    return CHISUBMIT_SUCCESS



@click.command(name="pull-grading-branches")
@click.argument('assignment_id', type=str)
@click.option('--from-staging', is_flag=True)
@click.option('--from-students', is_flag=True)
@click.option('--only', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_pull_grading_branches(ctx, course, assignment_id, from_staging, from_students, only):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)
    
    teams_registrations = get_teams_registrations(course, assignment, only = only)
    teams = sorted(teams_registrations.keys(), key=operator.attrgetter("team_id"))

    for team in teams:
        registration = teams_registrations[team]
        print "Pulling grading branch for team %s... " % team.team_id
        gradingrepo_pull_grading_branch(ctx.obj['config'], course, team, registration, from_staging = from_staging, from_students = from_students)

    return CHISUBMIT_SUCCESS



@click.command(name="add-rubrics")
@click.argument('assignment_id', type=str)
@click.option('--commit', is_flag=True)
@click.option('--all-teams', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_add_rubrics(ctx, course, assignment_id, commit, all_teams):
    assignment = course.get_assignment(assignment_id)
    if assignment is None:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)

    teams_registrations = get_teams_registrations(course, assignment, only_ready_for_grading=not all_teams)
    teams = sorted(teams_registrations.keys(), key=operator.attrgetter("team_id"))

    for team in teams:
        registration = teams_registrations[team]
        repo = GradingGitRepo.get_grading_repo(ctx.obj['config'], course, team, registration)
        
        if repo is None:
            print "%s does not have a grading repository" % team.team_id
            continue        
        
        rubric = RubricFile.from_assignment(assignment, registration.get_grades())
        rubricfile = "%s.rubric.txt" % assignment.assignment_id
        rubricfilepath = "%s/%s" % (repo.repo_path, rubricfile)
        
        if commit:
            if not os.path.exists(rubricfilepath):
                rubric.save(rubricfilepath, include_blank_comments=True)
                rv = repo.commit([rubricfile], "Added grading rubric")
                if rv:
                    print rubricfilepath, "(COMMITTED)"
                else:
                    print rubricfilepath, "(NO CHANGES - Not committed)"
            else:
                print rubricfilepath, "(SKIPPED - already exists)"
        else:
            rubric.save(rubricfilepath, include_blank_comments=True)
            print rubricfilepath


@click.command(name="collect-rubrics")
@click.argument('assignment_id', type=str)
@click.option('--dry-run', is_flag=True)
@click.option('--only', type=str)
@click.option('--grader-id', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grading_collect_rubrics(ctx, course, assignment_id, dry_run, only, grader_id):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    if grader_id is not None:
        grader = get_grader_or_exit(ctx, course, grader_id)
    else:
        grader = None

    rcs = assignment.get_rubric_components()

    teams_registrations = get_teams_registrations(course, assignment, grader=grader, only=only)
    teams = sorted(teams_registrations.keys(), key=operator.attrgetter("team_id"))
    
    for team in teams:
        registration = teams_registrations[team]
        repo = GradingGitRepo.get_grading_repo(ctx.obj['config'], course, team, registration)
        if repo is None:
            print "Repository for %s does not exist" % (team.team_id)
            continue

        rubricfile = repo.repo_path + "/%s.rubric.txt" % assignment.assignment_id

        if not os.path.exists(rubricfile):
            print "Repository for %s does not have a rubric for assignment %s" % (team.team_id, assignment.assignment_id)
            continue

        try:
            rubric = RubricFile.from_file(open(rubricfile), assignment)
        except ChisubmitRubricException, cre:
            print "ERROR: Rubric for %s does not validate (%s)" % (team.team_id, cre.message)
            continue

        points = []
        for rc in rcs:
            grade = rubric.points[rc.description]
            if grade is None:
                points.append(0.0)
            else:
                if not dry_run:
                    registration.set_grade(rc, grade)
                points.append(grade)

        adjustments = {}
        total_penalties = 0.0
        total_bonuses = 0.0
        if rubric.penalties is not None:
            for desc, p in rubric.penalties.items():
                adjustments[desc] = p
                total_penalties += p

        if rubric.bonuses is not None:
            for desc, p in rubric.bonuses.items():
                adjustments[desc] = p
                total_bonuses += p

        if not dry_run:
            registration.grade_adjustments = adjustments

        if ctx.obj["verbose"]:
            print team.team_id
            print "Points Obtained: %s" % points
            print "Penalties: %.2f" % total_penalties
            print "Bonuses: %.2f" % total_bonuses

        if not dry_run:
            total_grade = registration.get_total_grade()
        else:
            total_grade = sum(points) + total_penalties + total_bonuses
            
        if ctx.obj["verbose"]:
            print "TOTAL: %.2f" % total_grade
            print
        else:
            print "%-40s %.2f" % (team.team_id, total_grade)
            

instructor_grading.add_command(instructor_grading_set_grade)
instructor_grading.add_command(instructor_grading_load_grades)
instructor_grading.add_command(instructor_grading_list_grades)
instructor_grading.add_command(instructor_grading_assign_graders)
instructor_grading.add_command(instructor_grading_list_grader_assignments)
instructor_grading.add_command(instructor_grading_list_submissions)
instructor_grading.add_command(instructor_grading_show_grading_status)
instructor_grading.add_command(instructor_grading_create_grading_repos)
instructor_grading.add_command(instructor_grading_create_grading_branches)
instructor_grading.add_command(instructor_grading_push_grading_branches)
instructor_grading.add_command(instructor_grading_pull_grading_branches)
instructor_grading.add_command(instructor_grading_add_rubrics)
instructor_grading.add_command(instructor_grading_collect_rubrics)



from django.core.urlresolvers import reverse
from rest_framework import status
from rest_framework.test import APITestCase
from django.contrib.auth.models import User
from chisubmit.common.utils import get_datetime_now_utc
from datetime import timedelta
from chisubmit.backend.api.models import Assignment

class SubmitTests(APITestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                         'course1_pa1', 'course1_pa1_registrations']  
        
    def test_correct_no_extensions(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                    }
        response = self.client.post(url, data = post_data)
        response_data = response.json()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response_data["in_grace_period"], False)      
        self.assertEqual(response_data["submission"]["extensions_used"], 0)               
        self.assertEqual(response_data["extensions_before"], 2)
        self.assertEqual(response_data["extensions_after"], 2)
        
    def test_correct_no_extensions_grace_period(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        deadline = get_datetime_now_utc() - timedelta(minutes=10)
        assignment_obj = Assignment.objects.get(assignment_id = "pa1")
        assignment_obj.deadline = deadline
        assignment_obj.grace_period = timedelta(minutes=15)
        assignment_obj.save()

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                    }
        response = self.client.post(url, data = post_data)
        response_data = response.json()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response_data["in_grace_period"], True)
        self.assertEqual(response_data["submission"]["extensions_used"], 0)               
        self.assertEqual(response_data["extensions_before"], 2)
        self.assertEqual(response_data["extensions_after"], 2)
        
    def test_correct_dry_run(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                    }
        response = self.client.post(url + "?dry_run=true", data = post_data)
        response_data = response.json()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response_data["in_grace_period"], False)
        self.assertEqual(response_data["submission"]["extensions_used"], 0)                 
        
    def test_correct_one_extension(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)
        
        deadline = get_datetime_now_utc() - timedelta(hours=23)
        assignment_obj = Assignment.objects.get(assignment_id = "pa1")
        assignment_obj.deadline = deadline
        assignment_obj.save()

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                     "extensions": 1                   
                    }
        response = self.client.post(url, data = post_data)
        response_data = response.json()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response_data["in_grace_period"], False)    
        self.assertEqual(response_data["submission"]["extensions_used"], 1) 
        self.assertEqual(response_data["extensions_before"], 2)
        self.assertEqual(response_data["extensions_after"], 1)
        
    def test_correct_one_extension_grace_period(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)
        
        deadline = get_datetime_now_utc() - timedelta(hours=24, minutes=10)
        assignment_obj = Assignment.objects.get(assignment_id = "pa1")
        assignment_obj.deadline = deadline
        assignment_obj.grace_period = timedelta(minutes=15)        
        assignment_obj.save()

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                    }
        response = self.client.post(url, data = post_data)
        response_data = response.json()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response_data["in_grace_period"], True)
        self.assertEqual(response_data["submission"]["extensions_used"], 1)
        self.assertEqual(response_data["extensions_before"], 2)
        self.assertEqual(response_data["extensions_after"], 1)
        
    def test_incorrect_insufficient_extensions_in_team(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)
        
        deadline = get_datetime_now_utc() - timedelta(hours=23 + 24 + 24)
        assignment_obj = Assignment.objects.get(assignment_id = "pa1")
        assignment_obj.deadline = deadline
        assignment_obj.save()

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                    }
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        
    def test_incorrect_insufficient_extensions_in_team_extensions_override(self):
        user = User.objects.get(username='instructor1')
        self.client.force_authenticate(user=user)
        
        deadline = get_datetime_now_utc() - timedelta(hours=23 + 24 + 24)
        assignment_obj = Assignment.objects.get(assignment_id = "pa1")
        assignment_obj.deadline = deadline
        assignment_obj.save()

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                     "extensions_override": 2            
                    }
        response = self.client.post(url, data = post_data)
        response_data = response.json()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response_data["in_grace_period"], False)
        self.assertEqual(response_data["submission"]["extensions_used"], 2)                           
        self.assertEqual(response_data["extensions_override"], 2)        
        self.assertEqual(response_data["extensions_before"], 2)
        self.assertEqual(response_data["extensions_after"], 0)
        
class SubmitWithExistingSubmissionsTests(APITestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                         'course1_pa1', 'course1_pa1_registrations_with_submissions']
        
                


# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-16 16:22
from __future__ import unicode_literals

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0002_connstr_length'),
    ]

    operations = [
        migrations.AddField(
            model_name='assignment',
            name='grace_period',
            field=models.DurationField(default=datetime.timedelta(0)),
        ),
        migrations.AddField(
            model_name='submission',
            name='in_grace_period',
            field=models.BooleanField(default=False),
        ),
    ]



#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

from requests import exceptions, Session
from urlparse import urlparse
from pprint import pprint
import json
import sys
from requests.exceptions import HTTPError
from chisubmit.client.exceptions import UnknownObjectException,\
    ChisubmitRequestException, BadRequestException, UnauthorizedException
import base64
import datetime

def json_serial(obj):
    if isinstance(obj, datetime.timedelta):
        return str(obj)
    
    raise TypeError("Type not serializable")

class Requester(object):
    
    def __init__(self, login_or_token, password, base_url):
        
        self.__base_url = base_url
        
        self.__headers = {}
        self.__headers['content-type'] = 'application/json'
        if login_or_token is not None and password is not None:
            self.__headers["Authorization"] = "Basic " + base64.b64encode('%s:%s' % (login_or_token, password))
        elif login_or_token is not None:
            self.__headers["Authorization"] = "Token %s" % login_or_token
        
        self.__session = Session()

    def request(self, method, resource, data=None, headers=None, params=None):
        if resource.startswith("/"):
            url = self.__base_url + resource
        else:
            # TODO: Validate the URL is valid given base_url
            url = resource

        all_headers = {}
        all_headers.update(self.__headers)
        if headers is not None:
            all_headers.update(headers)

        if data is not None:
            data = json.dumps(data, default=json_serial)
            
        # TODO: try..except
        response = self.__session.request(url = url,
                                  method = method,
                                  params = params,
                                  data = data,
                                  headers = all_headers)
        
        if response.status_code == 400:
            raise BadRequestException(method, url, params, data, all_headers, response)        
        elif 400 < response.status_code < 500:
            if response.status_code == 401:
                raise UnauthorizedException(method, url, params, data, all_headers, response)
            if response.status_code == 404:
                raise UnknownObjectException(method, url, params, data, all_headers, response)
            else:
                raise ChisubmitRequestException(method, url, params, data, all_headers, response)
        elif 500 <= response.status_code < 600:
            raise ChisubmitRequestException(method, url, params, data, all_headers, response)

        try:
            response_data = response.json()
        except ValueError:
            response_data = {"data": response.text}

        return response.headers, response_data



from chisubmit.tests.common import cli_test, ChisubmitCLITestCase
from chisubmit.common.utils import get_datetime_now_utc, set_testing_now
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL

from datetime import timedelta
from chisubmit.backend.api.models import Course

class CLICompleteWorkflowGracePeriods(ChisubmitCLITestCase):
            
    fixtures = ['admin_user']
            
    @cli_test
    def test_complete_with_submission_grace_period(self, runner):
        course_id = u"cmsc40300"
        course_name = u"Foobarmentals of Foobar II"

        admin_id = u"admin"
        instructor_ids = [u"instructor"]
        grader_ids= [u"grader"]
        student_ids = [u"student1", u"student2", u"student3", u"student4"]
        
        all_users = instructor_ids + grader_ids + student_ids
        
        admin, instructors, graders, students = self.create_clients(runner, admin_id, instructor_ids, grader_ids, student_ids, course_id, verbose = True)
        self.create_users(admin, all_users)
        
        self.create_course(admin, course_id, course_name)
        
        course = Course.get_by_course_id(course_id)
        self.assertIsNotNone(course)
        self.assertEquals(course.name, course_name)        

        result = admin.run("admin course set-attribute %s default_extensions 2" % (course_id))
        self.assertEquals(result.exit_code, 0)
        
        result = admin.run("admin course set-attribute %s extension_policy per-student" % (course_id))
        self.assertEquals(result.exit_code, 0)
        
        self.add_users_to_course(admin, course_id, instructors, graders, students)
        
        deadline = get_datetime_now_utc() - timedelta(minutes=5)
        deadline = deadline.isoformat(sep=" ")

        result = instructors[0].run("instructor assignment add", 
                                    ["pa1", "Programming Assignment 1", deadline])
        self.assertEquals(result.exit_code, 0)
        
        result = instructors[0].run("instructor assignment set-attribute", 
                                    ["pa1", "max_students", "2"])
        self.assertEquals(result.exit_code, 0)        
        
        result = instructors[0].run("instructor assignment set-attribute", 
                                    ["pa1", "grace_period", "00:15"])
        self.assertEquals(result.exit_code, 0)           
            
        teams = [u"student1-student2", 
                 u"student3-student4"]        

        students_team = [ (students[0], students[1]),
                          (students[2], students[3])]
        

        self.register_team(students_team[0], teams[0], "pa1", course_id)
        self.register_team(students_team[1], teams[1], "pa1", course_id)
        
        _, _, team_commits = self.create_team_repos(admin, course_id, teams[0:2], students_team[0:2])
                
        # Team 0 submits during the grace period, and doesn't have to use an extension. 
        result = students_team[0][0].run("student assignment submit", 
                                         ["pa1", "--yes"])        
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)
        
        t = course.get_team(teams[0])
        
        self.assertEqual(t.get_extensions_available(), 2)
        
        
        # Let the grace period "pass"
        new_now = get_datetime_now_utc() + timedelta(minutes=11)
        set_testing_now(new_now)

        print
        print "~~~ Time has moved 'forward' to one minute after the grace period hours ~~~"
        print
        
        # Team 1 submits and has to use an extensions because the grace period is over
        result = students_team[1][0].run("student assignment submit", 
                                         ["pa1", "--yes"])        
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        t = course.get_team(teams[1])
        
        self.assertEqual(t.get_extensions_available(), 1)

        for team, student_team in zip(teams, students_team):
            result = student_team[0].run("student team show", [team])
            self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        
        



import tempfile
import os
import yaml
import sys
import colorama
import git
import functools
import re
from rest_framework.test import APILiveServerTestCase

from click.testing import CliRunner
from chisubmit.client.exceptions import BadRequestException
from chisubmit.cli import chisubmit_cmd
from chisubmit.backend.api.models import Course, Student
from django.contrib.auth.models import User
from rest_framework.authtoken.models import Token
from django.conf import settings

cli_verbose = False

def set_cli_verbose(v):
    global cli_verbose
    
    cli_verbose = v

colorama.init()

COURSE1_ID = "cmsc40100"
COURSE1_NAME = "Introduction to Software Testing"
COURSE1_INSTRUCTORS = ["instructor1"]
COURSE1_GRADERS =     ["grader1", "grader2"]
COURSE1_STUDENTS =    ["student1", "student2", "student3", "student4"]
COURSE1_USERS = COURSE1_INSTRUCTORS + COURSE1_GRADERS + COURSE1_STUDENTS
COURSE1_ASSIGNMENTS = ["pa1", "pa2"]
COURSE1_RUBRICS = {"pa1": ["First Task", "Second Task"],
                   "pa2": ["First Task", "Second Task", "Third Task"],}
COURSE1_TEAMS = ["student1-student2", "student3-student4"]
COURSE1_TEAM_MEMBERS = {"student1-student2": ["student1", "student2"],
                        "student3-student4": ["student3", "student4"]}

COURSE2_ID = "cmsc40110"
COURSE2_NAME = "Advanced Software Testing"
COURSE2_INSTRUCTORS = ["instructor2"]
COURSE2_GRADERS =     ["grader3", "grader4"]
COURSE2_STUDENTS =    ["student5", "student6", "student7", "student8"]
COURSE2_USERS = COURSE2_INSTRUCTORS + COURSE2_GRADERS + COURSE2_STUDENTS
COURSE2_ASSIGNMENTS = ["hw1", "hw2"]
    
ALL_USERS = list(set(COURSE1_USERS + COURSE2_USERS + ["admin"]))


def cli_test(func=None, isolated_filesystem = True):
    if func is None:
        return functools.partial(cli_test, isolated_filesystem = isolated_filesystem)
    def new_func(self, *args, **kwargs):
        runner = CliRunner()
        
        try:
            if isolated_filesystem:
                with runner.isolated_filesystem():
                    func(self, runner, *args, **kwargs)
            else:
                func(self, runner, *args, **kwargs)
        except BadRequestException, bre:
            bre.print_errors()
            raise
    return new_func


class ChisubmitCLITestClient(object):
    
    def __init__(self, api_url, user_id, api_key, runner, 
                       course = None, git_credentials = {}, verbose = False):
        global cli_verbose
        
        self.user_id = user_id
        self.home_dir = "test-fs/home/%s" % self.user_id
        self.work_dir = "%s/chisubmit-test" % (self.home_dir)
        self.conf_dir = "%s/.chisubmit" % (self.work_dir)
        self.conf_file = "%s/chisubmit.conf" % (self.conf_dir)

        self.runner = runner
        if cli_verbose:
            self.verbose = True
        else:
            self.verbose = verbose
        self.course = course
        
        git_credentials.update({"Testing" : "testing-credentials"})

        os.makedirs(self.home_dir)
        os.mkdir(self.work_dir)
        os.mkdir(self.conf_dir)
        with open(self.conf_file, 'w') as f:
            conf = {"api-url": api_url,
                    "api-key": api_key,
                    "git-credentials":
                        git_credentials
                    }
            yaml.safe_dump(conf, f, default_flow_style=False)   
            
    def run(self, subcommands, params = [], course = None, cmd=chisubmit_cmd, catch_exceptions=False, cmd_input = None):
        chisubmit_args = ['--debug', '--work-dir', self.work_dir, '--config-dir', self.conf_dir]
        
        if course is not None:
            chisubmit_args += ['-c', "course={}".format(course)]
        elif self.course is not None:
            chisubmit_args += ['-c', "course={}".format(self.course)]
        
        if subcommands is None:
            cmd_args = []
        else:
            cmd_args = subcommands.split()

        cmd_args += params

        if self.verbose:
            global cli_verbose
            if cli_verbose:
                print
            l = []
            for ca in cmd_args:
                if " " in ca:
                    l.append('"%s"' % ca)
                else:
                    l.append(ca)
            s = ""
            s+= colorama.Style.BRIGHT + colorama.Fore.BLUE
            s+= "%s$" % self.user_id
            s+= colorama.Fore.WHITE
            s+= " chisubmit %s" % " ".join(l)
            s+= colorama.Style.RESET_ALL
            print s
        
        result = self.runner.invoke(cmd, chisubmit_args + cmd_args, catch_exceptions=catch_exceptions, input=cmd_input)
        
        if self.verbose and len(result.output) > 0:
            sys.stdout.write(result.output)
            sys.stdout.flush()
            
        return result
    
    def create_local_git_repository(self, path):
        git_path = "%s/%s" % (self.home_dir, path)
                
        repo = git.Repo.init(git_path)

        return repo, git_path

    def get_local_git_repository(self, path):
        git_path = "%s/%s" % (self.home_dir, path)
                
        repo = git.Repo(git_path)

        return repo, git_path


class ChisubmitCLITestCase(APILiveServerTestCase):
    
    def __init__(self, *args, **kwargs):
        super(ChisubmitCLITestCase, self).__init__(*args, **kwargs)

        if settings.DEBUG == False:
            settings.DEBUG = True

        self.git_server_connstr = "server_type=Testing;local_path=./test-fs/server"
        self.git_staging_connstr = "server_type=Testing;local_path=./test-fs/staging"
        self.git_server_user = None
        self.git_staging_user = None
        self.git_api_keys = {}
    
    def set_git_server_connstr(self, connstr):
        self.git_server_connstr = connstr

    def set_git_server_user(self, user):
        self.git_server_user = user

    def set_git_staging_connstr(self, connstr):
        self.git_staging_connstr = connstr

    def set_git_staging_user(self, user):
        self.git_staging_user = user
        
    def add_api_key(self, git_type, apikey):
        self.git_api_keys[git_type] = apikey
    
    def assert_http_code(self, response, expected):
        self.assertEquals(response.status_code, expected, "Expected HTTP response code %i, got %i" % (expected, response.status_code))                
        
    def create_user(self, admin_runner, username):
        fname = "f_" + username
        lname = "l_" + username
        email = username + "@example.org"
        result = admin_runner.run("admin user add", [username, fname, lname, email])
        self.assertEquals(result.exit_code, 0)
            
        try:
            user_obj = User.objects.get(username=username)
        except User.DoesNotExist:
            self.fail("User was not added to database")              

        self.assertEquals(user_obj.username, username)            
        self.assertEquals(user_obj.first_name, fname)            
        self.assertEquals(user_obj.last_name, lname)            
        self.assertEquals(user_obj.email, email)      
        
        # Creating a custom token cannot be done throught the API or CLI, so
        # we create a Token object directly.
        Token.objects.create(key=username+"token", user=user_obj)

    def create_clients(self, runner, admin_id, instructor_ids = [], grader_ids = [], student_ids = [], course_id=None, verbose=False):
        base_url = self.live_server_url + "/api/v1"
        
        admin = ChisubmitCLITestClient(base_url, admin_id, admin_id+"token", runner, git_credentials=self.git_api_keys, verbose = verbose)
        
        instructors = []
        for instructor_id in instructor_ids:
            instructors.append(ChisubmitCLITestClient(base_url, instructor_id, instructor_id+"token", runner, git_credentials=self.git_api_keys,  verbose = verbose, course = course_id))

        graders = []
        for grader_id in grader_ids:
            graders.append(ChisubmitCLITestClient(base_url, grader_id, grader_id+"token", runner, git_credentials=self.git_api_keys,  verbose = verbose, course = course_id))

        students = []
        for student_id in student_ids:
            students.append(ChisubmitCLITestClient(base_url, student_id, student_id+"token", runner, git_credentials=self.git_api_keys,  verbose = verbose, course = course_id))
                        
        return admin, instructors, graders, students    
            
    def create_users(self, admin, user_ids):
        for user_id in user_ids:
            self.create_user(admin, user_id)

    def create_course(self, admin, course_id, course_name):
        result = admin.run("admin course add", [course_id, course_name])
        self.assertEquals(result.exit_code, 0)
        
        course = Course.get_by_course_id(course_id)
        self.assertIsNotNone(course)
        self.assertEquals(course.name, course_name)
        
        result = admin.run("admin course list")
        self.assertEquals(result.exit_code, 0)
        self.assertIn(course_id, result.output)
        self.assertIn(course_name, result.output)

        result = admin.run("admin course set-attribute %s git_usernames custom" % (course_id))
        self.assertEquals(result.exit_code, 0)
        
        git_server_connstr = self.git_server_connstr
        git_staging_connstr = self.git_staging_connstr

        result = admin.run("admin course set-attribute %s git_server_connstr %s" % (course_id, git_server_connstr))
        self.assertEquals(result.exit_code, 0)

        result = admin.run("admin course set-attribute %s git_staging_connstr %s" % (course_id, git_staging_connstr))
        self.assertEquals(result.exit_code, 0)
        
        course = Course.get_by_course_id(course_id)
        self.assertEquals(course.git_server_connstr, git_server_connstr)
        self.assertEquals(course.git_staging_connstr, git_staging_connstr)
        
        result = admin.run("admin course unsetup-repo %s" % (course_id))
        self.assertEquals(result.exit_code, 0)
        
        result = admin.run("admin course setup-repo %s" % (course_id))
        self.assertEquals(result.exit_code, 0)

        result = admin.run("admin course unsetup-repo %s --staging" % (course_id))
        self.assertEquals(result.exit_code, 0)

        result = admin.run("admin course setup-repo %s --staging" % (course_id))
        self.assertEquals(result.exit_code, 0)        
        
        
    def add_users_to_course(self, admin, course_id, instructors, graders, students):  
        for instructor in instructors:
            result = admin.run("admin course add-instructor %s %s" % (course_id, instructor.user_id))
            self.assertEquals(result.exit_code, 0)
        
        for grader in graders:
            result = admin.run("admin course add-grader %s %s" % (course_id, grader.user_id))
            self.assertEquals(result.exit_code, 0)
        
        for student in students:
            result = admin.run("admin course add-student %s %s" % (course_id, student.user_id))
            self.assertEquals(result.exit_code, 0)
                
        for instructor in instructors:
            if self.git_server_user is None:
                git_username = "git-" + instructors[0].user_id
            else:
                git_username = self.git_server_user
    
            result = instructors[0].run("instructor course set-user-attribute", 
                                    ["instructor", instructors[0].user_id, "git_username", git_username])
            self.assertEquals(result.exit_code, 0)

            if self.git_staging_user is None:
                git_staging_username = "git-" + instructors[0].user_id
            else:
                git_staging_username = self.git_staging_user
    
            result = instructors[0].run("instructor course set-user-attribute", 
                                    ["instructor", instructors[0].user_id, "git_staging_username", git_staging_username])
            self.assertEquals(result.exit_code, 0)


        for grader in graders:
            if self.git_server_user is None:
                git_username = "git-" + graders[0].user_id
            else:
                git_username = self.git_server_user
    
            result = instructors[0].run("instructor course set-user-attribute", 
                                        ["grader", graders[0].user_id, "git_username", git_username])
            self.assertEquals(result.exit_code, 0)

            if self.git_staging_user is None:
                git_staging_username = "git-" + graders[0].user_id
            else:
                git_staging_username = self.git_staging_user
    
            result = instructors[0].run("instructor course set-user-attribute", 
                                    ["grader", graders[0].user_id, "git_staging_username", git_staging_username])
            self.assertEquals(result.exit_code, 0)

                
        result = admin.run("admin course update-repo-access", [course_id])
        self.assertEquals(result.exit_code, 0)
                
        result = admin.run("admin course update-repo-access", [course_id, "--staging"])
        self.assertEquals(result.exit_code, 0)
        
        for student in students:
            if self.git_server_user is None:
                git_username = "git-" + student.user_id
            else:
                git_username = self.git_server_user
                
            result = student.run("student course set-git-username", [git_username])
            self.assertEquals(result.exit_code, 0)
            
            student_obj = Student.objects.get(course__course_id = course_id, user__username = student.user_id)
                
            self.assertEquals(student_obj.git_username, git_username)


    def create_team_repos(self, admin, course_id, teams, students_team):
        
        team_git_paths = []
        team_git_repos = []
        team_commits = []
        
        for team, students in zip(teams, students_team):
            result = admin.run("admin course team-repo-remove", [course_id, team, "--ignore-non-existing"])
            self.assertEquals(result.exit_code, 0)
            result = admin.run("admin course team-repo-remove", ["--staging", course_id, team, "--ignore-non-existing"])
            self.assertEquals(result.exit_code, 0)

            result = admin.run("admin course team-repo-create", [course_id, team, "--public"])
            self.assertEquals(result.exit_code, 0)
            result = admin.run("admin course team-repo-create", ["--staging", course_id, team, "--public"])
            self.assertEquals(result.exit_code, 0)

        
            result = students[0].run("student team repo-check", [team])
            self.assertEquals(result.exit_code, 0)
            
            r = re.findall(r"^Repository URL: (.*)$", result.output, flags=re.MULTILINE)
            self.assertEquals(len(r), 1, "student team repo-check didn't produce the expected output")
            repo_url = r[0]

            team_git_repo, team_git_path = students[0].create_local_git_repository(team)
            team_remote = team_git_repo.create_remote("origin", repo_url)
            
            team_git_repos.append(team_git_repo)
            team_git_paths.append(team_git_path)

            commits = []
                        
            files = ["foo", "bar", "baz"]
            
            for f in files:
                open("%s/%s" % (team_git_path, f), "w").close()
                
            team_git_repo.index.add(files)
            team_git_repo.index.commit("First commit of %s" % team)
                
            commits.append(team_git_repo.heads.master.commit)
    
            team_remote.push("master")
            
            f = open("%s/foo" % (team_git_path), "w")
            f.write("Hello, team1!")
            f.close()
            
            team_git_repo.index.add(["foo"])
            team_git_repo.index.commit("Second commit of %s" % team)

            commits.append(team_git_repo.heads.master.commit)
            
            team_remote.push("master")
            
            team_commits.append(commits)
            
        for team1, students1 in zip(teams, students_team):
            for team2, students2 in zip(teams, students_team):
                if team1 != team2:
                    result = students1[0].run("student team repo-check", [team2])
                    self.assertNotEquals(result.exit_code, 0)
                    result = students2[0].run("student team repo-check", [team1])
                    self.assertNotEquals(result.exit_code, 0)
            
        return team_git_paths, team_git_repos, team_commits

            
    def register_team(self, student_clients, team_name, assignment_id, course_id):
        for s in student_clients:
            partners = [s2 for s2 in student_clients if s2!=s]
            partner_args = []
            for p in partners:
                partner_args += ["--partner", p.user_id]
        
            s.run("student assignment register", 
                  [ assignment_id ] + partner_args)

            s.run("student team show", [team_name])
        
        return 
        students_in_team = [User.from_id(s.user_id) for s in student_clients]
        
        ts = Team.find_teams_with_students(course_id, students_in_team)
        
        self.assertGreaterEqual(len(ts), 1)
        self.assertIn(team_name, [t.id for t in ts])
        
        team = [t for t in ts if t.id == team_name]
        self.assertEqual(len(team), 1)
        t = team[0]
        self.assertEquals(t.id, team_name)
        self.assertListEqual(sorted([s.id for s in students_in_team]), 
                             sorted([s.id for s in t.students]))
        for st in t.students_teams:
            self.assertEquals(st.status, StudentsTeams.STATUS_CONFIRMED)      
    


import click
from chisubmit.cli.common import pass_course, DATETIME, get_assignment_or_exit,\
    catch_chisubmit_exceptions, require_local_config, get_team_or_exit,\
    get_assignment_registration_or_exit
from chisubmit.client.assignment import Assignment
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from chisubmit.common.utils import convert_datetime_to_utc, create_connection,\
    convert_datetime_to_local
import operator
from chisubmit.cli.shared.assignment import shared_assignment_list,\
    shared_assignment_set_attribute
from chisubmit.client.exceptions import UnknownObjectException,\
    BadRequestException
from chisubmit.cli.student.assignment import print_commit


@click.group(name="assignment")
@click.pass_context
def instructor_assignment(ctx):
    pass


@click.command(name="add")
@click.argument('assignment_id', type=str)
@click.argument('name', type=str)
@click.argument('deadline', type=DATETIME)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_assignment_add(ctx, course, assignment_id, name, deadline):
    deadline = convert_datetime_to_utc(deadline)
    
    course.create_assignment(assignment_id, name, deadline)
    
    return CHISUBMIT_SUCCESS


@click.command(name="add-rubric-component")
@click.argument('assignment_id', type=str)
@click.argument('description', type=str)
@click.argument('points', type=float)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_assignment_add_rubric_component(ctx, course, assignment_id, description, points):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)
    assignment.create_rubric_component(description, points)

    return CHISUBMIT_SUCCESS


@click.command(name="register")
@click.argument('assignment_id', type=str)
@click.option('--student', type=str, multiple=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_assignment_register(ctx, course, assignment_id, student):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)
    
    assignment.register(students = student)
    
    return CHISUBMIT_SUCCESS


@click.command(name="submit")
@click.argument('team_id', type=str)    
@click.argument('assignment_id', type=str)
@click.argument('commit_sha', type=str)
@click.argument('extensions', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context  
def instructor_assignment_submit(ctx, course, team_id, assignment_id, commit_sha, extensions):
    team = get_team_or_exit(ctx, course, team_id)
    registration = get_assignment_registration_or_exit(ctx, team, assignment_id)
            
    conn = create_connection(course, ctx.obj['config'])
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)
    
    commit = conn.get_commit(course, team, commit_sha)
    
    if commit is None:
        print "Commit %s does not exist in repository" % commit_sha
        ctx.exit(CHISUBMIT_FAIL)
    
    if registration.final_submission is not None:
        prior_commit_sha = registration.final_submission.commit_sha
        prior_submitted_at_utc = registration.final_submission.submitted_at
        prior_submitted_at_local = convert_datetime_to_local(prior_submitted_at_utc)            
        
        submission_commit = conn.get_commit(course, team, prior_commit_sha)            
        
        if prior_commit_sha == commit_sha:
            print "The team has already submitted assignment %s" % registration.assignment.assignment_id
            print "They submitted the following commit on %s:" % prior_submitted_at_local
            print
            if submission_commit is None:
                print "WARNING: Previously submitted commit '%s' is not in the repository!" % prior_commit_sha
            else:
                print_commit(submission_commit)
            print
            print "You are trying to submit the same commit again (%s)" % prior_commit_sha
            print "If you want to re-submit, please specify a different commit"
            ctx.exit(CHISUBMIT_FAIL)
            
        print
        print "WARNING: This team has already submitted assignment %s and you" % registration.assignment.assignment_id 
        print "are about to overwrite the previous submission of the following commit:"
        print
        if submission_commit is None:
            print "WARNING: Previously submitted commit '%s' is not in the repository!" % prior_commit_sha
        else:
            print_commit(submission_commit)
        print

    if registration.final_submission is not None:
        msg = "THE ABOVE SUBMISSION FOR %s (%s) WILL BE CANCELLED." % (registration.assignment.assignment_id, registration.assignment.name)
        
        print "!"*len(msg)
        print msg
        print "!"*len(msg)
        print
        print "If you continue, their submission for %s (%s)" % (registration.assignment.assignment_id, registration.assignment.name)
        print "will now point to the following commit:"                
    else:
        print "You are going to make a submission for %s (%s)." % (registration.assignment.assignment_id, registration.assignment.name)
        print "The commit you are submitting is the following:"
    print
    print_commit(commit)
    print
    print "PLEASE VERIFY THIS IS THE EXACT COMMIT YOU WANT TO SUBMIT"
    print
    print "Are you sure you want to continue? (y/n): ", 
    
    yesno = raw_input()
    
    if yesno in ('y', 'Y', 'yes', 'Yes', 'YES'):
        try:
            submit_response = registration.submit(commit_sha, extensions, ignore_deadline=True, dry_run=False)
                        
            print "The submission has been completed."
            return CHISUBMIT_SUCCESS

        except BadRequestException, bre:        
            print "ERROR: The submission was not completed. The server reported the following errors:"
            bre.print_errors()
            return CHISUBMIT_FAIL        



@click.command(name="stats")
@click.argument('assignment_id', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_assignment_stats(ctx, course, assignment_id):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)
             
    all_students = course.get_students()
         
    student_dict = {s.user.username: s.user for s in all_students if not s.dropped}
    students = set(student_dict.keys())
    dropped = set([s.user.username for s in all_students if s.dropped])
    
    teams_unconfirmed = []
    nteams = 0
    nstudents = len(student_dict)
    nstudents_assignment = 0
    nsubmissions = 0
    nstudents_submitted = 0
    unsubmitted = []
    
    for team in course.get_teams():
        try:
            registration = team.get_assignment_registration(assignment.assignment_id)
        except UnknownObjectException:
            continue
        
        unconfirmed = False
        includes_dropped = False
        team_members = team.get_team_members()
        for tm in team_members:
            if tm.username not in dropped:
                try:
                    students.remove(tm.username)
                    nstudents_assignment += 1
                    if not tm.confirmed:
                        unconfirmed = True
                except KeyError, ke:
                    print "WARNING: Student %s seems to be in more than one team (offending team: %s)" % (tm.username, team.team_id)
            else:
                includes_dropped = True

        if not includes_dropped:
            nteams += 1
            
            if registration.final_submission is not None:
                nsubmissions += 1
                nstudents_submitted += len(team_members)
            else:
                unsubmitted.append(team)
            
        if unconfirmed:
            teams_unconfirmed.append(team)
                
    assert(nstudents == len(students) + nstudents_assignment)                
            
    title = "Assignment '%s'" % (assignment.name)
    print title
    print "=" * len(title)
               
    print 
    print "%i / %i students in %i teams have signed up for assignment %s" % (nstudents_assignment, nstudents, nteams, assignment.assignment_id)
    print
    print "%i / %i teams have submitted the assignment (%i students)" % (nsubmissions, nteams, nstudents_submitted)
    
    if ctx.obj["verbose"]:
        if len(students) > 0:
            print
            print "Students who have not yet signed up"
            print "-----------------------------------"
            not_signed_up = [student_dict[sid] for sid in students]
            not_signed_up.sort(key=operator.attrgetter("last_name"))
            for s in not_signed_up:
                print "%s, %s <%s>" % (s.last_name, s.first_name, s.email)
                
        if len(unsubmitted) > 0:
            print
            print "Teams that have not submitted"
            print "-----------------------------"
            for t in unsubmitted:
                print t.team_id
                
    return CHISUBMIT_SUCCESS


instructor_assignment.add_command(shared_assignment_list)
instructor_assignment.add_command(shared_assignment_set_attribute)

instructor_assignment.add_command(instructor_assignment_add)
instructor_assignment.add_command(instructor_assignment_add_rubric_component)
instructor_assignment.add_command(instructor_assignment_register)
instructor_assignment.add_command(instructor_assignment_submit)
instructor_assignment.add_command(instructor_assignment_stats)




from chisubmit.backend.api.models import Course, TeamMember
from chisubmit.tests.integration.clientlibs import ChisubmitClientLibsTestCase
from chisubmit.tests.common import COURSE1_GRADERS, COURSE1_STUDENTS, COURSE1_INSTRUCTORS

class CoursePersonTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_users']
    
    def test_get_instructors(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        instructors = course.get_instructors()
        
        self.assertEquals(len(instructors), len(COURSE1_INSTRUCTORS))
        self.assertItemsEqual([i.username for i in instructors], COURSE1_INSTRUCTORS)
        
    def test_add_instructor_by_username(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40100")
        self.assertFalse(course_obj.instructors.filter(username="instructor2").exists())
        
        course = c.get_course("cmsc40100")
        instructor = course.add_instructor("instructor2")
        
        self.assertEquals(instructor.username, "instructor2")
        self.assertEquals(instructor.user.first_name, "F_instructor2")
        self.assertEquals(instructor.user.last_name, "L_instructor2")

        self.assertTrue(course_obj.instructors.filter(username="instructor2").exists())      
        
    def test_add_instructor_by_user(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40100")
        self.assertFalse(course_obj.instructors.filter(username="instructor2").exists())
        
        course = c.get_course("cmsc40100")
        user = c.get_user("instructor2")
        instructor = course.add_instructor(user)
        
        self.assertEquals(instructor.username, "instructor2")
        self.assertEquals(instructor.user.first_name, "F_instructor2")
        self.assertEquals(instructor.user.last_name, "L_instructor2")

        self.assertTrue(course_obj.instructors.filter(username="instructor2").exists())        
        
    def test_remove_instructor(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40100")
        self.assertTrue(course_obj.instructors.filter(username="instructor1").exists())
        
        course = c.get_course("cmsc40100")
        course.remove_instructor("instructor1")
        
        self.assertFalse(course_obj.instructors.filter(username="instructor1").exists())
        
    def test_get_graders(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        graders = course.get_graders()
        
        self.assertEquals(len(graders), len(COURSE1_GRADERS))
        self.assertItemsEqual([g.username for g in graders], COURSE1_GRADERS)
                
    def test_add_grader_by_username(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40100")
        self.assertFalse(course_obj.graders.filter(username="grader3").exists())
        
        course = c.get_course("cmsc40100")
        grader = course.add_grader("grader3")
        
        self.assertEquals(grader.username, "grader3")
        self.assertEquals(grader.user.first_name, "F_grader3")
        self.assertEquals(grader.user.last_name, "L_grader3")

        self.assertTrue(course_obj.graders.filter(username="grader3").exists())      
        
    def test_add_grader_by_user(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40100")
        self.assertFalse(course_obj.graders.filter(username="grader3").exists())
        
        course = c.get_course("cmsc40100")
        user = c.get_user("grader3")
        grader = course.add_grader(user)
        
        self.assertEquals(grader.username, "grader3")
        self.assertEquals(grader.user.first_name, "F_grader3")
        self.assertEquals(grader.user.last_name, "L_grader3")

        self.assertTrue(course_obj.graders.filter(username="grader3").exists())                   
                
    def test_remove_grader(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40100")
        self.assertTrue(course_obj.graders.filter(username="grader1").exists())
        
        course = c.get_course("cmsc40100")
        course.remove_grader("grader1")
        
        self.assertFalse(course_obj.graders.filter(username="grader1").exists())                
                
    def test_get_students(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        students = course.get_students()
        
        self.assertEquals(len(students), len(COURSE1_STUDENTS))
        self.assertItemsEqual([s.username for s in students], COURSE1_STUDENTS)      
        
    def test_add_student_by_username(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40100")
        self.assertFalse(course_obj.students.filter(username="student5").exists())
        
        course = c.get_course("cmsc40100")
        student = course.add_student("student5")
        
        self.assertEquals(student.username, "student5")
        self.assertEquals(student.user.first_name, "F_student5")
        self.assertEquals(student.user.last_name, "L_student5")

        self.assertTrue(course_obj.students.filter(username="student5").exists())      
        
    def test_add_student_by_user(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40100")
        self.assertFalse(course_obj.students.filter(username="student5").exists())
        
        course = c.get_course("cmsc40100")
        user = c.get_user("student5")
        student = course.add_student(user)
        
        self.assertEquals(student.username, "student5")
        self.assertEquals(student.user.first_name, "F_student5")
        self.assertEquals(student.user.last_name, "L_student5")

        self.assertTrue(course_obj.students.filter(username="student5").exists())    
        
    def test_remove_student(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40100")
        self.assertTrue(course_obj.students.filter(username="student1").exists())
        
        course = c.get_course("cmsc40100")
        course.remove_student("student1")
        
        self.assertFalse(course_obj.students.filter(username="student1").exists())   
        
class MultiCoursePersonTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 'course2']        
                
    def test_add_instructor(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40110")
        self.assertFalse(course_obj.instructors.filter(username="instructor1").exists())
        
        course = c.get_course("cmsc40110")
        instructor = course.add_instructor("instructor1")
        
        self.assertEquals(instructor.username, "instructor1")
        self.assertEquals(instructor.user.first_name, "F_instructor1")
        self.assertEquals(instructor.user.last_name, "L_instructor1")

        self.assertTrue(course_obj.instructors.filter(username="instructor1").exists())       
        
    def test_add_grader(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40110")
        self.assertFalse(course_obj.graders.filter(username="grader1").exists())
        
        course = c.get_course("cmsc40110")
        grader = course.add_grader("grader1")
        
        self.assertEquals(grader.username, "grader1")
        self.assertEquals(grader.user.first_name, "F_grader1")
        self.assertEquals(grader.user.last_name, "L_grader1")

        self.assertTrue(course_obj.graders.filter(username="grader1").exists())             
        
    def test_add_student(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40110")
        self.assertFalse(course_obj.students.filter(username="student1").exists())
        
        course = c.get_course("cmsc40110")
        student = course.add_student("student1")
        
        self.assertEquals(student.username, "student1")
        self.assertEquals(student.user.first_name, "F_student1")
        self.assertEquals(student.user.last_name, "L_student1")

        self.assertTrue(course_obj.students.filter(username="student1").exists())
        
    def test_add_team(self):
        c = self.get_api_client("admintoken")

        course_obj = Course.objects.get(course_id="cmsc40110")
        self.assertFalse(course_obj.students.filter(username="student1").exists())
        
        course = c.get_course("cmsc40110")
        course.add_student("student1")
        course.add_student("student2")
        
        team = course.create_team(team_id = "student1-student2",
                                  extensions = 2,
                                  active = True)
        self.assertEquals(team.team_id, "student1-student2")
        self.assertEquals(team.extensions, 2)
        self.assertEquals(team.active, True)
               
        team_obj = course_obj.get_team("student1-student2")
        self.assertIsNotNone(team_obj, "Team was not added to database")
            
        self.assertEquals(team_obj.team_id, "student1-student2")                  
        self.assertEquals(team_obj.extensions, 2)                  
        self.assertEquals(team_obj.active, True)
              
        team_member = team.add_team_member("student1", confirmed = True)
        self.assertEqual(team_member.username, "student1")
        self.assertEqual(team_member.student.user.username, "student1")
        self.assertEqual(team_member.confirmed, True)
                
        team_member_objs = TeamMember.objects.filter(team = team_obj, student__user__username = "student1")
        self.assertEquals(len(team_member_objs), 1)
        team_member_obj = team_member_objs[0]
        self.assertEqual(team_member_obj.student.user.username, "student1")
        self.assertEqual(team_member_obj.confirmed, True)

        team_member = team.add_team_member("student2", confirmed = True)
        self.assertEqual(team_member.username, "student2")
        self.assertEqual(team_member.student.user.username, "student2")
        self.assertEqual(team_member.confirmed, True)
                
        team_member_objs = TeamMember.objects.filter(team = team_obj, student__user__username = "student2")
        self.assertEquals(len(team_member_objs), 1)
        team_member_obj = team_member_objs[0]
        self.assertEqual(team_member_obj.student.user.username, "student2")
        self.assertEqual(team_member_obj.confirmed, True)
        
        team_members = team.get_team_members()
        
        self.assertEquals(len(team_members), 2)
        self.assertItemsEqual([tm.username for tm in team_members], ["student1","student2"])
        self.assertItemsEqual([tm.student.user.username for tm in team_members], ["student1","student2"])             
        


import click
from chisubmit.client.assignment import Assignment
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from chisubmit.common.utils import convert_datetime_to_local,\
    create_connection, get_datetime_now_utc, compute_extensions_needed,\
    is_submission_ready_for_grading
from dateutil.parser import parse
from chisubmit.cli.common import pass_course, get_assignment_or_exit,\
    get_team_or_exit, get_assignment_registration_or_exit,\
    catch_chisubmit_exceptions, require_local_config,\
    get_team_registration_from_user
from chisubmit.cli.shared.assignment import shared_assignment_list
from datetime import timedelta
from chisubmit.client.exceptions import BadRequestException,\
    UnknownObjectException


@click.group(name="assignment")
@click.pass_context
def student_assignment(ctx):
    pass


@click.command(name="register")
@click.argument('assignment_id', type=str)
@click.option('--partner', type=str, multiple=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def student_assignment_register(ctx, course, assignment_id, partner):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)
    
    user = ctx.obj["client"].get_user()
    
    try:
        course.get_instructor(user.username)
        
        # If get_instructor doesn't raise an exception, then the user
        # is an instructor, and we don't include the user in the list
        # of partners.
        r = assignment.register(students = partner)    
    except UnknownObjectException:
        # Otherwise, we include the current user
        if user.username in partner:
            print "You specified your own username in --partner. You should use this"
            print "option to specify your partners, not including yourself."
            return CHISUBMIT_FAIL
        
        r = assignment.register(students = partner + (user.username,))    

    print "Your registration for %s (%s) is complete." % (r.registration.assignment.assignment_id, r.registration.assignment.name)
    
    if len(r.team_members) > 1:
        some_unconfirmed = False
        print 
        print "Your team name is '%s'." % r.team.team_id
        print
        print "The team is composed of the following students:"
        print
        for tm in r.team_members:
            if tm.confirmed:
                conf = ""
            else:
                conf = "UNCONFIRMED"
                some_unconfirmed = True
                
            print " - %s, %s (%s) %s" % (tm.student.user.last_name, 
                                         tm.student.user.first_name,
                                         tm.student.user.username,
                                         conf)
             
            if some_unconfirmed:
                print
                print "Note: Some students have not yet confirmed that they are part of"
                print "      this team. To confirm they are part of this team, they just"
                print "      need to register as a team themselves (using this same"
                print "      command, and listing the same team members)."
    
    return CHISUBMIT_SUCCESS


@click.command(name="cancel-registration")
@click.argument('assignment_id', type=str)
@click.option('--yes', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context  
def student_assignment_cancel_registration(ctx, course, assignment_id, yes):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    # Get registration
    team, registration = get_team_registration_from_user(ctx, course, assignment)
    team_members = team.get_team_members()  
    
    if len(team_members) == 1:
        student = team_members[0].student        
        individual = True
        print "You (%s %s) currently have an INDIVIDUAL registration for %s (%s)" % (student.user.first_name, student.user.last_name, assignment.assignment_id, assignment.name)
    else:
        students = [tm.student for tm in team_members]
        individual = False
        print "You currently have a TEAM registration for %s (%s)." % (assignment.assignment_id, assignment.name)
        print
        print "Team %s has the following students:" % team.team_id
        for s in students:
            print " - %s %s" % (s.user.first_name, s.user.last_name)         
    
    print
    
    if registration.final_submission is not None:
        if individual:
            print "You have already made a submission for %s." % assignment_id
        else:
            print "Team %s has already made a submission for assignment %s." % (team.team_id, assignment_id)
        print
        print "Please cancel your submission first, and then try canceling your registration."
        ctx.exit(CHISUBMIT_FAIL)

    if individual:
        print "Are you sure you want to cancel your registration for this assignment? (y/n): ",
    else:
        print "Are you sure you want to cancel this team's registration for the assignment? (y/n): ",
     
    
    if not yes:
        yesno = raw_input()
    else:
        yesno = 'y'
        print 'y'
    
    if yesno in ('y', 'Y', 'yes', 'Yes', 'YES'):        
        try:
            registration.cancel()
            print
            print "Your registration has been cancelled."
            ctx.exit(CHISUBMIT_SUCCESS)
        except BadRequestException, bre:
            response_data = bre.json
            print "ERROR: Your registration cannot be cancelled. The server reported the following:"
            print
            bre.print_errors()

            ctx.exit(CHISUBMIT_FAIL)


@click.command(name="show-deadline")
@click.argument('assignment_id', type=str)
@click.option('--utc', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def student_assignment_show_deadline(ctx, course, assignment_id, utc):
    assignment = course.get_assignment(assignment_id)
    if assignment is None:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)

    now_utc = get_datetime_now_utc()
    now_local = convert_datetime_to_local(now_utc)

    deadline_utc = assignment.deadline
    deadline_local = convert_datetime_to_local(deadline_utc)

    print assignment.name
    print
    if utc:
        print "      Now (Local): %s" % now_local.isoformat(" ")
        print " Deadline (Local): %s" % deadline_local.isoformat(" ")
        print
        print "        Now (UTC): %s" % now_utc.isoformat(" ")
        print "   Deadline (UTC): %s" % deadline_utc.isoformat(" ")
    else:
        print "      Now: %s" % now_local.isoformat(" ")
        print " Deadline: %s" % deadline_local.isoformat(" ")

    print

    extensions = compute_extensions_needed(now_utc, deadline_utc)

    if extensions == 0:
        diff = deadline_utc - now_utc
    else:
        diff = now_utc - deadline_utc

    days = diff.days
    hours = diff.seconds // 3600
    minutes = (diff.seconds//60)%60
    seconds = diff.seconds%60

    if extensions == 0:
        print "The deadline has not yet passed"
        print "You have %i days, %i hours, %i minutes, %i seconds left" % (days, hours, minutes, seconds)
    else:
        print "The deadline passed %i days, %i hours, %i minutes, %i seconds ago" % (days, hours, minutes, seconds)
        print "If you submit your assignment now, you will need to use %i extensions" % extensions

    return CHISUBMIT_SUCCESS


def print_commit(commit):
    print "      Commit: %s" % commit.sha
    print "        Date: %s" % commit.committed_date.isoformat(sep=" ")
    print "     Message: %s" % commit.message
    print "      Author: %s <%s>" % (commit.author_name, commit.author_email)        


@click.command(name="submit")
@click.argument('assignment_id', type=str)
@click.option('--commit-sha', type=str)
@click.option('--yes', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context  
def student_assignment_submit(ctx, course, assignment_id, commit_sha, yes):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    # Determine team for this assignment
    team, registration = get_team_registration_from_user(ctx, course, assignment)
    team_members = team.get_team_members()    
                
    title = "SUBMISSION FOR ASSIGNMENT %s (%s)" % (assignment.assignment_id, assignment.name)
    print title
    print "-" * len(title)
    print
    if len(team_members) == 1:
        student = team_members[0].student
        individual = True
        print "This is an INDIVIDUAL submission for %s %s" % (student.user.first_name, student.user.last_name)
    else:
        students = [tm.student for tm in team_members]
        individual = False
        print "This is a TEAM submission for team %s with the following students:" % team.team_id
        for s in students:
            print " - %s %s" % (s.user.first_name, s.user.last_name)
    print
                
    conn = create_connection(course, ctx.obj['config'])
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)
    
    if commit_sha is None:
        commit = conn.get_latest_commit(course, team)
                
        user_specified_commit = False
    else:    
        commit = conn.get_commit(course, team, commit_sha)
        
        if commit is None:
            print "Commit %s does not exist in repository" % commit_sha
            ctx.exit(CHISUBMIT_FAIL)

        user_specified_commit = True
        
    try:
        submit_response = registration.submit(commit.sha, dry_run = True)
    except BadRequestException, bre:
        response_data = bre.json
        
        if "extensions_needed" in response_data and "extensions_available" in response_data:
            extensions_needed = response_data["extensions_needed"]
            extensions_available = response_data["extensions_available"]
            
            deadline_utc = parse(response_data["deadline"])
            submitted_at_utc = parse(response_data["submitted_at"])
            deadline_local = convert_datetime_to_local(deadline_utc)
            submitted_at_local = convert_datetime_to_local(submitted_at_utc)        
            
            if extensions_needed > extensions_available:
                msg1 = "You do not have enough extensions to submit this assignment."
                msg2 = "You would need %i extensions to submit this assignment at this " \
                       "time, but you only have %i left" % (extensions_needed, extensions_available)
    
                print
                print msg1
                print            
                print "     Deadline (UTC): %s" % deadline_utc.isoformat(sep=" ")
                print "          Now (UTC): %s" % submitted_at_utc.isoformat(sep=" ")
                print 
                print "   Deadline (Local): %s" % deadline_local.isoformat(sep=" ")
                print "        Now (Local): %s" % submitted_at_local.isoformat(sep=" ")
                print 
                print msg2 
                print
            else:
                print "ERROR: Your submission cannot be completed. The server reported the following:"
                print
                bre.print_errors()
        else:
            print "ERROR: Your submission cannot be completed. The server reported the following:"
            print
            bre.print_errors()

        ctx.exit(CHISUBMIT_FAIL)
    
    if registration.final_submission is not None:
        prior_commit_sha = registration.final_submission.commit_sha
        prior_extensions_used = registration.final_submission.extensions_used             
        prior_submitted_at_utc = registration.final_submission.submitted_at
        prior_submitted_at_local = convert_datetime_to_local(prior_submitted_at_utc)            
        
        submission_commit = conn.get_commit(course, team, prior_commit_sha)            
        
        if prior_commit_sha == commit.sha:
            print "You have already submitted assignment %s" % registration.assignment.assignment_id
            print
            print "You submitted the following commit on %s:" % prior_submitted_at_local
            print
            if submission_commit is None:
                print "WARNING: Previously submitted commit '%s' is not in the repository!" % prior_commit_sha
            else:
                print_commit(submission_commit)
            print
            if user_specified_commit:
                print "You are trying to submit the same commit again (%s)" % prior_commit_sha
                print "If you want to re-submit, please specify a different commit."
            else:
                print "The above commit is the latest commit in your repository."
                print
                print "If you were expecting to see a different commit, make sure you've pushed"
                print "your latest code to your repository."
            ctx.exit(CHISUBMIT_FAIL)
            
        print "You have already submitted assignment %s" % registration.assignment.assignment_id
        print
        print "You submitted the following commit on %s:" % prior_submitted_at_local
        print
        if submission_commit is None:
            print "WARNING: Previously submitted commit '%s' is not in the repository!" % prior_commit_sha
        else:
            print_commit(submission_commit)
        print

        msg = "IF YOU CONTINUE, THE ABOVE SUBMISSION FOR %s (%s) WILL BE CANCELLED." % (registration.assignment.assignment_id, registration.assignment.name)
        
        print "!"*len(msg)
        print msg
        print "!"*len(msg)
        print        
        if not user_specified_commit:
            print "If you continue, your submission will instead point to the latest commit in your repository:"
        else:
            print "If you continue, your submission will instead point to the following commit:"                
    else:
        if not user_specified_commit:
            print "The latest commit in your repository is the following:"
        else:
            print "The commit you are submitting is the following:"
    print
    print_commit(commit)
    print
    print "PLEASE VERIFY THIS IS THE EXACT COMMIT YOU WANT TO SUBMIT"
    print
    if individual:
        print "You currently have %i extensions" % (submit_response.extensions_before)
    else:
        print "Your team currently has %i extensions" % (submit_response.extensions_before)
    print
    if registration.final_submission is not None:
        print "You used %i extensions in your previous submission of this assignment." % prior_extensions_used
        print "and you are going to use %i additional extensions now." % (submit_response.extensions_needed - prior_extensions_used)
    else:
        print "You are going to use %i extensions on this submission." % submit_response.extensions_needed
    print
    print "You will have %i extensions left after this submission." % submit_response.extensions_after
    print
    
    if submit_response.in_grace_period:
        print "NOTE: You are submitting after the deadline, but the instructor has"
        print "allowed some extra time after the deadline for students to submit"
        print "without having to consume an extension."
        print
     
    print "Are you sure you want to continue? (y/n): ", 
    
    if not yes:
        yesno = raw_input()
    else:
        yesno = 'y'
        print 'y'
    
    if yesno in ('y', 'Y', 'yes', 'Yes', 'YES'):
        try:
            submit_response = registration.submit(commit.sha, dry_run=False)
            
            # TODO: Can't do this until GitLab supports updating tags
            #    
            # message = "Extensions: %i\n" % extensions_requested
            # if submission_tag is None:
            #     conn.create_submission_tag(course, team, tag_name, message, commit.sha)
            # else:
            #     conn.update_submission_tag(course, team, tag_name, message, commit.sha)
            
            print
            print "Your submission has been completed."

            if submit_response.in_grace_period:
                print
                print "Your submission was made during the deadline's grace period. This means"
                print "that, although your submission was technically made *after* the"
                print "deadline, we are counting it as if it had been made before the deadline."
                print
                print "In the future, you should not rely on the presence of this grace period!"
                print "Your instructor may choose not to use one in future assignments, or may"
                print "use a shorter grace period. Your instructor is also aware of what"
                print "submissions are made during the grace period; if you repeatedly submit"
                print "during the grace period, your instructor may bring this to your attention."
            
            return CHISUBMIT_SUCCESS

        except BadRequestException, bre:
            print        
            print "ERROR: Your submission was not completed. The server reported the following errors:"
            bre.print_errors()
            return CHISUBMIT_FAIL
    else:
        print "Your submission has not been completed."
        print
        print "If you chose not to proceed because the above commit is not the one you wanted"
        print "to submit, make sure you've pushed your latest code to your repository before"
        print "attempting to submit again."
        print
        print "If you want to submit a different commit from your latest commit (e.g., an earlier"
        print "commit), you can use the --commit-sha option to specify a different commit."
        return CHISUBMIT_FAIL
    
    
@click.command(name="cancel-submit")   
@click.argument('assignment_id', type=str)
@click.option('--yes', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context  
def student_assignment_cancel_submit(ctx, course, assignment_id, yes):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    # Determine team for this assignment
    team, registration = get_team_registration_from_user(ctx, course, assignment)
    team_members = team.get_team_members()    
    
    if len(team_members) == 1:
        individual = True
    else:
        individual = False    
    
    if registration.final_submission is None:
        if individual:
            print "You have not made a submission for assignment %s," % assignment_id
        else:
            print "Team %s has not made a submission for assignment %s," % (team.team_id, assignment_id)
        print "so there is nothing to cancel."
        ctx.exit(CHISUBMIT_FAIL)
        
    if is_submission_ready_for_grading(assignment_deadline=registration.assignment.deadline, 
                                       submission_date=registration.final_submission.submitted_at,
                                       extensions_used=registration.final_submission.extensions_used):
        print "You cannot cancel this submission."
        print "You made a submission before the deadline, and the deadline has passed."

        ctx.exit(CHISUBMIT_FAIL)        
            
    conn = create_connection(course, ctx.obj['config'])
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)
        
    submission_commit = conn.get_commit(course, team, registration.final_submission.commit_sha)
        
    print
    print "This is your existing submission for assignment %s:" % assignment_id
    print
    if submission_commit is None:
        print "WARNING: Previously submitted commit '%s' is not in the repository!" % registration.final_submission.commit_sha
    else:
        print_commit(submission_commit)
    print    

    print "Are you sure you want to cancel this submission? (y/n): ", 
    
    if not yes:
        yesno = raw_input()
    else:
        yesno = 'y'
        print 'y'
    
    if yesno in ('y', 'Y', 'yes', 'Yes', 'YES'):
        registration.final_submission_id = None
        
        # TODO: Can't do this until GitLab supports updating tags
        #    
        # message = "Extensions: %i\n" % extensions_requested
        # if submission_tag is None:
        #     conn.create_submission_tag(course, team, tag_name, message, commit.sha)
        # else:
        #     conn.update_submission_tag(course, team, tag_name, message, commit.sha)
        print
        print "Your submission has been cancelled."
        

student_assignment.add_command(shared_assignment_list)

student_assignment.add_command(student_assignment_register)
student_assignment.add_command(student_assignment_cancel_registration)
student_assignment.add_command(student_assignment_show_deadline)
student_assignment.add_command(student_assignment_submit)
student_assignment.add_command(student_assignment_cancel_submit)



#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

import chisubmit.client.course
from chisubmit.client.requester import Requester

class Chisubmit(object):
    
    def __init__(self, login_or_token, base_url, password = None, deferred_save = False):
        # TODO: Validate URL 
        
        self._requester = Requester(login_or_token, password, base_url.rstrip("/"))
        self._deferred_save = deferred_save
    
    def get_courses(self):
        """
        :calls: GET /courses/
        :rtype: List of :class:`chisubmit.client.course.Course`
        """
        
        headers, data = self._requester.request(
            "GET",
            "/courses/"
        )
        return [chisubmit.client.course.Course(self, headers, elem) for elem in data]    
    
    def get_course(self, course_id, include_users=False, include_assignments=False, include_teams=False):
        """
        :calls: GET /courses/:course
        :param course_id: string
        :rtype: :class:`chisubmit.client.course.Course`
        """
        assert isinstance(course_id, (str, unicode)), course_id
        
        include = []
        
        if include_users:
            include.append("instructors")
            include.append("students")
            include.append("graders")

        if include_assignments:
            include.append("assignments")
            include.append("assignments__rubric")
            
        if include_teams:
            include.append("teams")
            
        if len(include) > 0:
            params = {"include": include}
        else:
            params = None            
        
        headers, data = self._requester.request(
            "GET",
            "/courses/" + course_id,
            params = params
        )
        return chisubmit.client.course.Course(self, headers, data)
 
    def create_course(self, course_id, name, git_usernames = None, git_staging_usernames = None, 
                      extension_policy = None, default_extensions = None):
        """
        :calls: POST /courses/
        :param course_id: string
        :param name: string
        :param git_usernames: string
        :param git_staging_usernames: string
        :param extension_policy: string
        :param default_extensions: int
        :rtype: :class:`chisubmit.client.course.Course`
        """
        assert isinstance(course_id, (str, unicode)), course_id
        
        post_data = {"course_id": course_id,
                     "name": name}
        
        if git_usernames is not None:
            post_data["git_usernames"] = git_usernames
        if git_staging_usernames is not None:
            post_data["git_staging_usernames"] = git_staging_usernames
        if extension_policy is not None:
            post_data["extension_policy"] = extension_policy
        if default_extensions is not None:
            post_data["default_extensions"] = default_extensions
        
        headers, data = self._requester.request(
            "POST",
            "/courses/",
            data = post_data
        )
        return chisubmit.client.course.Course(self, headers, data)

    def get_users(self):
        """
        :calls: GET /users/
        :rtype: List of :class:`chisubmit.client.users.User`
        """
        
        headers, data = self._requester.request(
            "GET",
            "/users/"
        )
        return [chisubmit.client.users.User(self, headers, elem) for elem in data]    
    
    def get_user(self, username = None):
        """
        :calls: GET /users/:username or GET /user
        :param username: string
        :rtype: :class:`chisubmit.client.users.User`
        """
        assert isinstance(username, (str, unicode)) or username is None, username
        
        if username is None:
            headers, data = self._requester.request(
                "GET",
                "/user"
            )            
        else:        
            headers, data = self._requester.request(
                "GET",
                "/users/" + username
            )
        return chisubmit.client.users.User(self, headers, data)
    
    def get_user_token(self, username = None, reset=False):
        """
        :calls: GET /users/:username/token or GET /user/token
        :param username: string
        :rtype: token: string, created: bool
        """
        assert isinstance(username, (str, unicode)) or username is None, username
        
        if reset:
            params = {"reset":"true"}
        else:
            params = {}
        
        if username is None:
            headers, data = self._requester.request(
                "GET",
                "/user/token",
                params = params
            )            
        else:        
            headers, data = self._requester.request(
                "GET",
                "/users/" + username + "/token",
                params = params
            )
        
        return data["token"], data["new"]    
    
    def create_user(self, username, first_name, last_name, email):
        """
        :calls: POST /users/
        :param username: string
        :param first_name: string
        :param last_name: string
        :param email: string
        :rtype: :class:`chisubmit.client.users.User`
        """
        assert isinstance(username, (str, unicode)), username
        
        post_data = {"username": username,
                     "first_name": first_name,
                     "last_name": last_name,
                     "email": email}
        
        headers, data = self._requester.request(
            "POST",
            "/users/",
            data = post_data
        )
        return chisubmit.client.users.User(self, headers, data)    



from django.core.urlresolvers import reverse
from rest_framework import status
from rest_framework.test import APITestCase
from django.contrib.auth.models import User

from pprint import pprint
from chisubmit.backend.api.models import Course

class TeamTests(APITestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams']
    
    def test_get_teams_as_admin(self):
        user = User.objects.get(username='admin')
        self.client.force_authenticate(user=user)
        
        url = reverse('team-list', args=["cmsc40100"])
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_teams_as_instructor(self):
        user = User.objects.get(username='instructor1')
        self.client.force_authenticate(user=user)
        
        url = reverse('team-list', args=["cmsc40100"])
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        
    def test_get_teams_as_student(self):
        user = User.objects.get(username='instructor1')
        self.client.force_authenticate(user=user)
        
        url = reverse('team-list', args=["cmsc40100"])
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        


from rest_framework import serializers
from chisubmit.backend.api.models import Course, GradersAndStudents, AllExceptAdmin,\
    Students, Student, Instructor, Grader, Team, Assignment, ReadWrite,\
    OwnerPermissions, Read, RubricComponent, TeamMember, Registration,\
    Submission, Grade
from django.contrib.auth.models import User
from rest_framework.reverse import reverse
from rest_framework.relations import RelatedField
from django.core.exceptions import ObjectDoesNotExist
from django.utils.encoding import smart_text
from django.utils.translation import ugettext_lazy

class ChisubmitSerializer(serializers.Serializer):
    
    def to_representation(self, obj):
        # TODO: Avoid generating the representation for fields that
        # aren't going to be returned anyways
        data = super(ChisubmitSerializer, self).to_representation(obj)

        course = self.context.get("course", None)
        request = self.context.get("request", None)
        if request is not None:
            user = request.user
        else:
            user = None

        if course is not None and user is not None:
            is_owner = self.context.get("is_owner", False)
            owner_override = getattr(self, "owner_override", {})
            
            if hasattr(self, "hidden_fields"):
                roles = self.context.get("roles", set())
                fields = data.keys()
                for f in fields:
                    if f in self.hidden_fields:
                        if not (is_owner and OwnerPermissions.READ in owner_override.get(f, [])):
                            if roles.issubset(self.hidden_fields[f]):
                                data.pop(f)
        
        return data
    
    def to_internal_value(self, data):
        internal_value = super(ChisubmitSerializer, self).to_internal_value(data)

        course = self.context.get("course", None)
        request = self.context.get("request", None)
        if request is not None:
            user = request.user
        else:
            user = None
        is_owner = self.context.get("is_owner", False)
        
        if course is not None and user is not None:
            roles = self.context.get("roles", set())
            fields = internal_value.keys()
            owner_override = getattr(self, "owner_override", {})

            for f in fields:
                if hasattr(self, "hidden_fields") and f in self.hidden_fields:
                    if not (is_owner and OwnerPermissions.READ in owner_override.get(f, [])):
                        if roles.issubset(self.hidden_fields[f]):
                            internal_value.pop(f)
                elif hasattr(self, "readonly_fields") and f in self.readonly_fields:
                    if not (is_owner and OwnerPermissions.WRITE in owner_override.get(f, [])):
                        if roles.issubset(self.readonly_fields[f]):
                            internal_value.pop(f)

        return internal_value
          
        
class UserSerializer(serializers.Serializer):
    username = serializers.CharField(max_length=30)
    first_name = serializers.CharField(max_length=30)
    last_name = serializers.CharField(max_length=30)
    email = serializers.EmailField() 
    
    def create(self, validated_data):
        return User.objects.create(**validated_data)
    
    def update(self, instance, validated_data):
        instance.username = validated_data.get('username', instance.username)
        instance.first_name = validated_data.get('first_name', instance.first_name)
        instance.last_name = validated_data.get('last_name', instance.last_name)
        instance.email = validated_data.get('email', instance.email)
        instance.save()
        return instance      
    

class CourseSerializer(ChisubmitSerializer):
    course_id = serializers.SlugField()
    name = serializers.CharField(max_length=64)
    
    url = serializers.SerializerMethodField()    
    instructors_url = serializers.SerializerMethodField()
    graders_url = serializers.SerializerMethodField()
    students_url = serializers.SerializerMethodField()
    assignments_url = serializers.SerializerMethodField()
    teams_url = serializers.SerializerMethodField()
    
    
    git_server_connstr = serializers.CharField(max_length=256, required=False)
    git_staging_connstr = serializers.CharField(max_length=256, required=False)
    git_usernames = serializers.ChoiceField(choices=Course.GIT_USERNAME_CHOICES, default=Course.GIT_USERNAME_USER)
    git_staging_usernames = serializers.ChoiceField(choices=Course.GIT_USERNAME_CHOICES, default=Course.GIT_USERNAME_USER)
    extension_policy = serializers.ChoiceField(choices=Course.EXT_CHOICES, default=Course.EXT_PER_STUDENT)
    default_extensions = serializers.IntegerField(default=0, min_value=0)    
    
    hidden_fields = { "git_staging_connstr": Students,    
                      "git_usernames": GradersAndStudents,
                      "git_staging_usernames": GradersAndStudents,
                      "extension_policy": GradersAndStudents,
                      "default_extensions": GradersAndStudents
                    }
    
    readonly_fields = { "course_id": AllExceptAdmin,
                        "name": AllExceptAdmin,
                        "git_server_connstr": AllExceptAdmin,
                        "git_staging_connstr": AllExceptAdmin,                        
                        "git_usernames": AllExceptAdmin,
                        "git_staging_usernames": AllExceptAdmin,
                        "extension_policy": AllExceptAdmin,
                        "default_extensions": AllExceptAdmin
                      }

    def get_url(self, obj):
        return reverse('course-detail', args=[obj.course_id], request=self.context["request"])

    def get_instructors_url(self, obj):
        return reverse('instructor-list', args=[obj.course_id], request=self.context["request"])    

    def get_graders_url(self, obj):
        return reverse('grader-list', args=[obj.course_id], request=self.context["request"])    
    
    def get_students_url(self, obj):
        return reverse('student-list', args=[obj.course_id], request=self.context["request"])    
    
    def get_assignments_url(self, obj):
        return reverse('assignment-list', args=[obj.course_id], request=self.context["request"])      

    def get_teams_url(self, obj):
        return reverse('team-list', args=[obj.course_id], request=self.context["request"])      
    
    def create(self, validated_data):
        return Course.objects.create(**validated_data)

    def update(self, instance, validated_data):
        instance.course_id = validated_data.get('course_id', instance.course_id)
        instance.name = validated_data.get('name', instance.name)
        instance.git_server_connstr = validated_data.get('git_server_connstr', instance.git_server_connstr)
        instance.git_staging_connstr = validated_data.get('git_staging_connstr', instance.git_staging_connstr)
        instance.git_usernames = validated_data.get('git_usernames', instance.git_usernames)
        instance.git_staging_usernames = validated_data.get('git_staging_usernames', instance.git_staging_usernames)
        instance.extension_policy = validated_data.get('extension_policy', instance.extension_policy)
        instance.default_extensions = validated_data.get('default_extensions', instance.default_extensions)
        instance.save()
        return instance
    
    
class InstructorSerializer(ChisubmitSerializer):
    url = serializers.SerializerMethodField()
    username = serializers.SlugRelatedField(
        source="user",
        queryset=User.objects.all(),
        slug_field='username'
    )
    user = UserSerializer(read_only=True)
    git_username = serializers.CharField(max_length=64, required=False)
    git_staging_username = serializers.CharField(max_length=64, required=False)
    
    hidden_fields = { "git_username": AllExceptAdmin,
                      "git_staging_username": AllExceptAdmin }
    
    readonly_fields = { }
    
    owner_override = {"git_username": ReadWrite,
                      "git_staging_username": ReadWrite }
    
    def get_url(self, obj):
        return reverse('instructor-detail', args=[self.context["course"].course_id, obj.user.username], request=self.context["request"])
    
    def create(self, validated_data):
        return Instructor.objects.create(**validated_data)
    
    def update(self, instance, validated_data):
        instance.git_username = validated_data.get('git_username', instance.git_username)
        instance.git_staging_username = validated_data.get('git_staging_username', instance.git_staging_username)
        instance.save()
        return instance    
    

class GraderSerializer(ChisubmitSerializer):
    url = serializers.SerializerMethodField()
    username = serializers.SlugRelatedField(
        source="user",
        queryset=User.objects.all(),
        slug_field='username'
    )
    user = UserSerializer(read_only=True)
    git_username = serializers.CharField(max_length=64, required=False)
    git_staging_username = serializers.CharField(max_length=64, required=False)
    
    hidden_fields = { "git_username": GradersAndStudents,
                      "git_staging_username": GradersAndStudents }
    
    readonly_fields = { }     
    
    owner_override = {"git_username": ReadWrite,
                      "git_staging_username": ReadWrite }    
    
    def get_url(self, obj):
        return reverse('grader-detail', args=[self.context["course"].course_id, obj.user.username], request=self.context["request"])
    
    def create(self, validated_data):
        return Grader.objects.create(**validated_data)
    
    def update(self, instance, validated_data):
        instance.git_username = validated_data.get('git_username', instance.git_username)
        instance.git_staging_username = validated_data.get('git_staging_username', instance.git_staging_username)
        instance.save()
        return instance       
    
    
class StudentSerializer(ChisubmitSerializer):
    url = serializers.SerializerMethodField()
    username = serializers.SlugRelatedField(
        source="user",
        queryset=User.objects.all(),
        slug_field='username'
    )
    user = UserSerializer(read_only=True, required=False)
    git_username = serializers.CharField(max_length=64, required=False)
    
    extensions = serializers.IntegerField(min_value=0, required=False)
    dropped = serializers.BooleanField(default=False)
    
    hidden_fields = { "git_username": Students, 
                      "dropped": Students }
    
    readonly_fields = { "extensions": GradersAndStudents,
                        "dropped": GradersAndStudents
                      }     
    
    owner_override = { "git_username": ReadWrite }
        
    def get_url(self, obj):
        return reverse('student-detail', args=[self.context["course"].course_id, obj.user.username], request=self.context["request"])
    
    def create(self, validated_data):
        if not validated_data.has_key("extensions") and self.context["course"].extension_policy == "per-student":
            validated_data["extensions"] = self.context["course"].default_extensions
      
        return Student.objects.create(**validated_data)
    
    def update(self, instance, validated_data):
        instance.git_username = validated_data.get('git_username', instance.git_username)
        instance.extensions = validated_data.get('extensions', instance.extensions)
        instance.dropped = validated_data.get('dropped', instance.dropped)
        instance.save()
        return instance    
    
    
    
class AssignmentSerializer(ChisubmitSerializer):
    assignment_id = serializers.SlugField()
    name = serializers.CharField(max_length=64)
    deadline = serializers.DateTimeField()
    grace_period = serializers.DurationField(required=False)
    
    url = serializers.SerializerMethodField()
    rubric_url = serializers.SerializerMethodField()
    
    min_students = serializers.IntegerField(default=1, min_value=1)
    max_students = serializers.IntegerField(default=1, min_value=1)
    
    hidden_fields = { "grace_period": GradersAndStudents }    
    
    readonly_fields = { "assignment_id": GradersAndStudents,
                        "name": GradersAndStudents,
                        "deadline": GradersAndStudents,
                        "min_students": GradersAndStudents,
                        "max_students": GradersAndStudents
                      }       
    
    def get_url(self, obj):
        return reverse('assignment-detail', args=[self.context["course"].course_id, obj.assignment_id], request=self.context["request"])

    def get_rubric_url(self, obj):
        return reverse('rubric-list', args=[self.context["course"].course_id, obj.assignment_id], request=self.context["request"])
    
    def create(self, validated_data):
        return Assignment.objects.create(**validated_data)
    
    def update(self, instance, validated_data):
        instance.assignment_id = validated_data.get('assignment_id', instance.assignment_id)
        instance.name = validated_data.get('name', instance.name)
        instance.deadline = validated_data.get('deadline', instance.deadline)
        instance.grace_period = validated_data.get('grace_period', instance.grace_period)
        instance.min_students = validated_data.get('min_students', instance.min_students)
        instance.max_students = validated_data.get('max_students', instance.max_students)
        instance.save()
        return instance    
  

class RubricComponentSerializer(ChisubmitSerializer):
    id = serializers.IntegerField(read_only = True)
    order = serializers.IntegerField(default=1, min_value=1)
    description = serializers.CharField(max_length=64)
    points = serializers.DecimalField(max_digits=5, decimal_places=2)
    
    url = serializers.SerializerMethodField()   
    
    readonly_fields = { "order": GradersAndStudents,
                        "description": GradersAndStudents,
                        "points": GradersAndStudents
                      }       
    
    def get_url(self, obj):
        return reverse('rubric-detail', args=[self.context["course"].course_id, obj.assignment.assignment_id, obj.pk], request=self.context["request"])
    
    def create(self, validated_data):
        return RubricComponent.objects.create(**validated_data)
    
    def update(self, instance, validated_data):
        instance.order = validated_data.get('order', instance.order)
        instance.description = validated_data.get('description', instance.description)
        instance.points = validated_data.get('points', instance.points)
        instance.save()
        return instance

    
class TeamSerializer(ChisubmitSerializer):
    team_id = serializers.SlugField()
    extensions = serializers.IntegerField(default=0, min_value=0)
    active = serializers.BooleanField(default = True)
        
    url = serializers.SerializerMethodField()
    students_url = serializers.SerializerMethodField()
    assignments_url = serializers.SerializerMethodField()

    readonly_fields = { "active": GradersAndStudents,
                        "team_id": GradersAndStudents,
                        "extensions": GradersAndStudents
                      }       
    
    def get_url(self, obj):
        return reverse('team-detail', args=[self.context["course"].course_id, obj.team_id], request=self.context["request"])

    def get_students_url(self, obj):
        return reverse('teammember-list', args=[self.context["course"].course_id, obj.team_id], request=self.context["request"])

    def get_assignments_url(self, obj):
        return reverse('registration-list', args=[self.context["course"].course_id, obj.team_id], request=self.context["request"])
    
    def create(self, validated_data):
        return Team.objects.create(**validated_data)
    
    def update(self, instance, validated_data):
        instance.team_id = validated_data.get('team_id', instance.team_id)
        instance.extensions = validated_data.get('extensions', instance.extensions)
        instance.active = validated_data.get('active', instance.min_students)
        instance.save()
        return instance         
    
class PersonRelatedField(RelatedField):
    default_error_messages = {
        'does_not_exist': ugettext_lazy('Object with username={value} does not exist.'),
        'invalid': ugettext_lazy('Invalid value.'),
    }
    
    def __init__(self, **kwargs):    
        super(PersonRelatedField, self).__init__(**kwargs)

    def to_internal_value(self, data):
        try:
            return self.get_queryset().get(user__username = data)
        except ObjectDoesNotExist:
            self.fail('does_not_exist', value=smart_text(data))
        except (TypeError, ValueError):
            self.fail('invalid')

    def to_representation(self, obj):
        return obj.user.username

    
class TeamMemberSerializer(ChisubmitSerializer):    
    url = serializers.SerializerMethodField()
    username = PersonRelatedField(source = "student",
                                  queryset=Student.objects.all()
                                  )
    student = StudentSerializer(read_only=True)
    confirmed = serializers.BooleanField(default=False)
    
    readonly_fields = { "confirmed": GradersAndStudents }        

    def __init__(self, *args, **kwargs):
        if kwargs.has_key("context"):        
            username_f = self.fields['username']
            username_f.queryset = username_f.queryset.filter(course = kwargs['context']['course'])

        super(TeamMemberSerializer, self).__init__(*args, **kwargs)

    def get_url(self, obj):
        return reverse('teammember-detail', args=[self.context["course"].course_id, obj.team.team_id, obj.student.user.username], request=self.context["request"])
    
    def create(self, validated_data):
        return TeamMember.objects.create(**validated_data)
    
    def update(self, instance, validated_data):
        instance.confirmed = validated_data.get('confirmed', instance.confirmed)
        instance.save()
        return instance
    
class SubmissionSerializer(ChisubmitSerializer):    
    url = serializers.SerializerMethodField()
    id = serializers.IntegerField(read_only = True)
    extensions_used = serializers.IntegerField(default=0, min_value=0) 
    commit_sha = serializers.CharField(max_length=40)
    submitted_at = serializers.DateTimeField(required=False)
    
    readonly_fields = { 
                      "extensions_used": GradersAndStudents,
                      "commit_sha": GradersAndStudents,
                      "submitted_at": GradersAndStudents
                      }   

    def get_url(self, obj):
        if obj.pk is None:
            # If we do a dry-run submission, we will be seralizing a Submission object that has not
            # been saved, and thus doesn't have a primary key (or an endpoint)
            return None
        else:
            return reverse('submission-detail', args=[self.context["course"].course_id, obj.registration.team.team_id, obj.registration.assignment.assignment_id, obj.pk], request=self.context["request"])
    
    def create(self, validated_data):
        return Submission.objects.create(**validated_data)
    
    def update(self, instance, validated_data):
        instance.extensions_used = validated_data.get('extensions_used', instance.extensions_used)
        instance.commit_sha = validated_data.get('commit_sha', instance.commit_sha)
        instance.submitted_at = validated_data.get('submitted_at', instance.submitted_at)
        instance.save()        
        return instance    
    
class RegistrationSerializer(ChisubmitSerializer):    
    url = serializers.SerializerMethodField()
    assignment_id = serializers.SlugRelatedField(
        source="assignment",
        queryset=Assignment.objects.all(),
        slug_field='assignment_id'
    )
    assignment = AssignmentSerializer(read_only=True, required=False) 
    grader_username = PersonRelatedField(
                                         source = "grader",
                                         queryset=Grader.objects.all(),
                                         required = False,
                                         allow_null=True
                                         )
    grader = GraderSerializer(read_only=True, required=False)

    submissions_url = serializers.SerializerMethodField()
    final_submission_id = serializers.PrimaryKeyRelatedField(
        source="final_submission",
        queryset=Submission.objects.all(),
        required=False,
        allow_null=True
        )
    final_submission = SubmissionSerializer(read_only=True, required=False) 

    grades_url = serializers.SerializerMethodField()
    grade_adjustments = serializers.DictField(required=False,
                                              child=serializers.DecimalField(max_digits=5, decimal_places=2))

    readonly_fields = { "grade_adjustments": GradersAndStudents }

    hidden_fields = { 
                      "grader_username": Students,
                      "grader": Students                      
                    }   

    def __init__(self, *args, **kwargs):
        if kwargs.has_key("context"):   
            grader_username_f = self.fields['grader_username']
            grader_username_f.queryset = grader_username_f.queryset.filter(course = kwargs['context']['course'])

        super(RegistrationSerializer, self).__init__(*args, **kwargs)

    def get_url(self, obj):
        return reverse('registration-detail', args=[self.context["course"].course_id, obj.team.team_id, obj.assignment.assignment_id], request=self.context["request"])

    def get_submissions_url(self, obj):
        return reverse('submission-list', args=[self.context["course"].course_id, obj.team.team_id, obj.assignment.assignment_id], request=self.context["request"])

    def get_grades_url(self, obj):
        return reverse('grade-list', args=[self.context["course"].course_id, obj.team.team_id, obj.assignment.assignment_id], request=self.context["request"])
    
    def create(self, validated_data):
        return Registration.objects.create(**validated_data)
    
    def update(self, instance, validated_data):
        instance.grader = validated_data.get('grader', instance.grader)
        instance.grade_adjustments = validated_data.get('grade_adjustments', instance.grade_adjustments)
        instance.final_submission = validated_data.get('final_submission', instance.final_submission)
        instance.save()
        return instance            
    

class RegistrationRequestSerializer(serializers.Serializer):
    students = serializers.ListField(
                                      child = serializers.CharField()
                                      )
    
class RegistrationResponseSerializer(serializers.Serializer):
    new_team = serializers.BooleanField()
    team = TeamSerializer()
    team_members = serializers.ListField(
                                         child = TeamMemberSerializer()
                                         )
    registration = RegistrationSerializer()    
    
class SubmissionRequestSerializer(serializers.Serializer):
    commit_sha = serializers.CharField(max_length=40)
    extensions_override = serializers.IntegerField(required = False, allow_null = True, min_value=0)

class SubmissionResponseSerializer(serializers.Serializer):
    submission = SubmissionSerializer()
    extensions_before = serializers.IntegerField(min_value=0) 
    extensions_needed = serializers.IntegerField(min_value=0) 
    extensions_after = serializers.IntegerField(min_value=0) 
    extensions_override = serializers.IntegerField(required = False)     
    in_grace_period = serializers.BooleanField()
    
class GradeSerializer(ChisubmitSerializer):
    rubric_component_id = serializers.PrimaryKeyRelatedField(
        source="rubric_component",
        queryset=RubricComponent.objects.all()
    )    
    rubric_component = RubricComponentSerializer(read_only=True)
    points = serializers.DecimalField(max_digits=5, decimal_places=2, required = False)
    
    url = serializers.SerializerMethodField()
        
    readonly_fields = { "points": GradersAndStudents }       
    
    def get_url(self, obj):
        return reverse('grade-detail', args=[self.context["course"].course_id, obj.registration.team.team_id, obj.registration.assignment.assignment_id, obj.pk], request=self.context["request"])
        
    def create(self, validated_data):
        return Grade.objects.create(**validated_data)
    
    def update(self, instance, validated_data):
        instance.points = validated_data.get('points', instance.points)
        instance.save()
        return instance    
    
          
  


import click
from chisubmit.cli.common import pass_course, get_student_or_exit,\
    api_obj_set_attribute, catch_chisubmit_exceptions, require_local_config
from chisubmit.cli.shared.course import shared_course_list,\
    shared_course_get_git_credentials
from chisubmit.client.exceptions import UnknownObjectException
from chisubmit.common import CHISUBMIT_FAIL, CHISUBMIT_SUCCESS

@click.group(name="course")
@click.pass_context
def student_course(ctx):
    pass

@click.command(name="show-extensions")
@click.option('--username', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def student_course_show_extensions(ctx, course, username):
    user = ctx.obj["client"].get_user()
    
    if username is not None:
        try:
            course.get_instructor(user.username)
        except UnknownObjectException:
            print "Only instructors can use the --username option"
            ctx.exit(CHISUBMIT_FAIL)
    else:
        username = user.username
   
    student = get_student_or_exit(ctx, course, username)
    
    if course.extension_policy == "per-team":
        print "This course uses per-team extensions."
        print "Please use 'chisubmit student team show' to see the number of extensions"
        print "available for the teams you are in."
        ctx.exit(CHISUBMIT_SUCCESS)
        
    print "%s, %s" % (student.user.last_name, student.user.first_name) 
    print
    teams = course.get_teams()
    
    extensions = student.extensions
    extensions_used = []
    
    for t in teams:
        if username not in [tm.username for tm in t.get_team_members()]:
            continue
        
        registrations = t.get_assignment_registrations()
        for reg in registrations:
            if reg.final_submission is not None:
                ex = reg.final_submission.extensions_used
                if ex > 0:
                    extensions_used.append((reg.assignment_id, t.team_id, ex))
                    
    n_extensions_used = sum([x[2] for x in extensions_used])
                    
    print "You started with %i extensions and you have %i extensions left." % (extensions, extensions - n_extensions_used)
    
    if len(extensions_used) > 0:
        for aid, tid, ex in extensions_used:
            print "- You used %i extension(s) on %s (as '%s')" % (ex, aid, tid)
                
    
        

@click.command(name="set-git-username")
@click.argument('git-username', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def student_course_set_git_username(ctx, course, git_username):
    user = ctx.obj["client"].get_user()
    
    student = get_student_or_exit(ctx, course, user.username)
    
    api_obj_set_attribute(ctx, student, "git_username", git_username)    
    

student_course.add_command(shared_course_list)
student_course.add_command(student_course_show_extensions)
student_course.add_command(student_course_set_git_username)
student_course.add_command(shared_course_get_git_credentials)




from django.http.response import Http404
from django.contrib.auth.models import User
from chisubmit.backend.api.models import Assignment, Team, TeamMember, Course,\
    CourseRoles, RubricComponent, Registration, Submission, Grade

def get_course(request, course_id):
    try:
        course_obj = Course.objects.get(course_id=course_id)
    except Course.DoesNotExist:
        raise Http404                           
    
    roles = course_obj.get_roles(request.user)
    
    if len(roles) == 0:
        raise Http404
    
    return course_obj, roles


def get_course_person(course_obj, request_user, roles, course_user_class, username):
    try:
        if len(roles) == 1 and CourseRoles.STUDENT in roles:
            if request_user.username != username:
                raise Http404
            
        user_obj = User.objects.get(username = username)
        person_obj = course_user_class.objects.get(course = course_obj, user = user_obj)
        return person_obj
    except User.DoesNotExist:
        raise Http404  
    except course_user_class.DoesNotExist:
        raise Http404 
    
def get_assignment(course_obj, request_user, roles, assignment_id):
    try:
        return Assignment.objects.get(course = course_obj, assignment_id = assignment_id)
    except Assignment.DoesNotExist:
        raise Http404  
    
def get_rubric_component(course_obj, request_user, roles, assignment_id, rubric_component_id):
    try:
        return RubricComponent.objects.get(assignment=assignment_id, pk=rubric_component_id)
    except RubricComponent.DoesNotExist:
        raise Http404      
    
def get_team(course_obj, request_user, roles, team_id):
    try:
        team_obj = Team.objects.get(course = course_obj, team_id = team_id)
        
        if len(roles) == 1 and CourseRoles.STUDENT in roles:
            if not team_obj.teammember_set.filter(student__user = request_user).exists():
                raise Http404
        
        return team_obj
    except Team.DoesNotExist:
        raise Http404
    
def get_team_member(course_obj, request_user, roles, team_id, student_username):
    try:
        team_obj = get_team(course_obj, request_user, roles, team_id)
        teammember_obj = TeamMember.objects.get(team = team_obj, student__user__username = student_username)
        return teammember_obj
    except (Team.DoesNotExist, TeamMember.DoesNotExist):
        raise Http404 
    
def get_registration(course_obj, request_user, roles, team_id, assignment_id):
    try:
        team_obj = get_team(course_obj, request_user, roles, team_id)
                
        if len(roles) == 1 and CourseRoles.STUDENT in roles:
            if not team_obj.teammember_set.filter(student__user = request_user).exists():
                raise Http404      
        
        registration_obj = Registration.objects.get(team = team_obj, assignment__assignment_id = assignment_id)
        return registration_obj
    except (Team.DoesNotExist, Registration.DoesNotExist):
        raise Http404         
    
def get_submission(course_obj, request_user, roles, team_id, assignment_id, submission_id):
    try:
        registration_obj = get_registration(course_obj, request_user, roles, team_id, assignment_id)
        submission_obj = Submission.objects.get(registration = registration_obj, pk = submission_id)
        
        return submission_obj
    except (Team.DoesNotExist, Registration.DoesNotExist, Submission.DoesNotExist):
        raise Http404       
    
def get_grade(course_obj, request_user, roles, team_id, assignment_id, grade_id):
    try:
        registration_obj = get_registration(course_obj, request_user, roles, team_id, assignment_id)
        grade_obj = Grade.objects.get(registration = registration_obj, pk = grade_id)
        
        return grade_obj
    except (Team.DoesNotExist, Registration.DoesNotExist, Grade.DoesNotExist):
        raise Http404                  


from rest_framework.test import APILiveServerTestCase
from chisubmit import client
    
class ChisubmitClientLibsTestCase(APILiveServerTestCase):
        
    def get_api_client(self, api_token, password=None, deferred_save = False):
        base_url = self.live_server_url + "/api/v1"
        
        return client.Chisubmit(login_or_token=api_token, password=password, base_url=base_url, deferred_save=deferred_save)  